from __future__ import annotations

import argparse
import json
from pathlib import Path

import pytest

from gedcom_tools.cli import create_parser, main
from gedcom_tools.commands.search import register_subcommand, run
from gedcom_tools.constants import EXIT_SUCCESS, EXIT_USAGE_ERROR

_FIXTURES = Path(__file__).parent / "fixtures"
_SAMPLE_GED = _FIXTURES / "555sample.ged"
_ROYAL_GED = _FIXTURES / "royal92.ged"


def _write_ged(tmp_path: Path, content: str, filename: str = "test.ged") -> Path:
    p = tmp_path / filename
    p.write_text(f"0 HEAD\n1 CHAR UTF-8\n{content}0 TRLR\n")
    return p


def _args(
    file: Path,
    query: str,
    *,
    regex: bool = False,
    fuzzy_dates: int | None = None,
    limit: int | None = None,
    count: bool = False,
    format: str = "text",
    verbose: bool = False,
    quiet: bool = False,
    no_color: bool = True,
    phonetic: str = "soundex",
) -> argparse.Namespace:
    return argparse.Namespace(
        file=file,
        query=query,
        regex=regex,
        fuzzy_dates=fuzzy_dates,
        limit=limit,
        count=count,
        format=format,
        verbose=verbose,
        quiet=quiet,
        no_color=no_color,
        phonetic=phonetic,
    )


_FAMILY_GED = (
    "0 @I1@ INDI\n1 NAME John /Smith/\n1 SEX M\n1 BIRT\n2 DATE 1800\n"
    "0 @I2@ INDI\n1 NAME Mary /Jones/\n1 SEX F\n1 BIRT\n2 DATE 1830\n"
    "0 @I3@ INDI\n1 NAME James /Smith/\n1 SEX M\n1 BIRT\n2 DATE 1860\n"
    "0 @F1@ FAM\n1 HUSB @I1@\n1 WIFE @I2@\n1 CHIL @I3@\n"
)


# ---------------------------------------------------------------------------
# CLI argument parsing
# ---------------------------------------------------------------------------


class TestArgumentParsing:
    def test_search_registered_as_subcommand(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        register_subcommand(subparsers)
        args = parser.parse_args(["search", "tree.ged", "Smith"])
        assert args.command == "search"

    def test_positional_file_parsed(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        register_subcommand(subparsers)
        args = parser.parse_args(["search", "myfile.ged", "Smith"])
        assert args.file == Path("myfile.ged")

    def test_positional_query_parsed(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        register_subcommand(subparsers)
        args = parser.parse_args(["search", "myfile.ged", "surname:Jones"])
        assert args.query == "surname:Jones"

    def test_regex_flag_defaults_false(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        register_subcommand(subparsers)
        args = parser.parse_args(["search", "f.ged", "Smith"])
        assert args.regex is False

    def test_regex_flag_set(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        register_subcommand(subparsers)
        args = parser.parse_args(["search", "f.ged", "Smith", "--regex"])
        assert args.regex is True

    def test_fuzzy_dates_parsed(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        register_subcommand(subparsers)
        args = parser.parse_args(["search", "f.ged", "born:1850", "--fuzzy-dates", "5"])
        assert args.fuzzy_dates == 5

    def test_fuzzy_dates_defaults_none(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        register_subcommand(subparsers)
        args = parser.parse_args(["search", "f.ged", "Smith"])
        assert args.fuzzy_dates is None

    def test_limit_parsed(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        register_subcommand(subparsers)
        args = parser.parse_args(["search", "f.ged", "Smith", "--limit", "10"])
        assert args.limit == 10

    def test_limit_defaults_none(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        register_subcommand(subparsers)
        args = parser.parse_args(["search", "f.ged", "Smith"])
        assert args.limit is None

    def test_count_flag_set(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        register_subcommand(subparsers)
        args = parser.parse_args(["search", "f.ged", "Smith", "--count"])
        assert args.count is True

    def test_count_defaults_false(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        register_subcommand(subparsers)
        args = parser.parse_args(["search", "f.ged", "Smith"])
        assert args.count is False

    def test_global_format_json_via_main_parser(self) -> None:
        parser = create_parser()
        args = parser.parse_args(["--format", "json", "search", "f.ged", "Smith"])
        assert args.format == "json"

    def test_global_verbose_via_main_parser(self) -> None:
        parser = create_parser()
        args = parser.parse_args(["--verbose", "search", "f.ged", "Smith"])
        assert args.verbose is True

    def test_global_quiet_via_main_parser(self) -> None:
        parser = create_parser()
        args = parser.parse_args(["--quiet", "search", "f.ged", "Smith"])
        assert args.quiet is True

    def test_global_no_color_via_main_parser(self) -> None:
        parser = create_parser()
        args = parser.parse_args(["--no-color", "search", "f.ged", "Smith"])
        assert args.no_color is True


# ---------------------------------------------------------------------------
# Full pipeline
# ---------------------------------------------------------------------------


class TestFullPipeline:
    def test_bare_name_search_finds_match(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME John /Smith/\n1 SEX M\n",
        )
        rc = run(_args(ged, "Smith"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "Smith" in out

    def test_surname_exact_match(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME John /Smith/\n"
            "0 @I2@ INDI\n1 NAME Anna /Smithson/\n",
        )
        rc = run(_args(ged, "surname=Smith"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "Smith" in out
        assert "Smithson" not in out

    def test_soundex_finds_similar_surname(self, tmp_path: Path, capsys) -> None:
        # Schmidt and Smith share soundex S530
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME Hans /Schmidt/\n1 SEX M\n",
        )
        rc = run(_args(ged, "surname~Smith"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "Schmidt" in out

    def test_metaphone_finds_cross_language_match(self, tmp_path: Path, capsys) -> None:
        # Catherine/Katherine differ in soundex initial letter but match in metaphone
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME Jane /Katherine/\n1 SEX F\n",
        )
        # Should NOT match with default soundex
        rc = run(_args(ged, "surname~Catherine"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "Katherine" not in out

        # Should match with metaphone
        rc = run(_args(ged, "surname~Catherine", phonetic="metaphone"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "Katherine" in out

    def test_date_range_search(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME Early /Person/\n1 BIRT\n2 DATE 1810\n"
            "0 @I2@ INDI\n1 NAME Late /Person/\n1 BIRT\n2 DATE 1900\n",
        )
        rc = run(_args(ged, "born:1800-1850"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "Early" in out
        assert "Late" not in out

    def test_place_substring_search(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME London /Person/\n"
            "1 BIRT\n2 PLAC London, England\n"
            "0 @I2@ INDI\n1 NAME Paris /Person/\n"
            "1 BIRT\n2 PLAC Paris, France\n",
        )
        rc = run(_args(ged, "place:England"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "London" in out
        assert "Paris" not in out

    def test_multiple_and_criteria_narrows_results(
        self, tmp_path: Path, capsys
    ) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME John /Smith/\n1 SEX M\n"
            "1 BIRT\n2 DATE 1850\n"
            "0 @I2@ INDI\n1 NAME Mary /Smith/\n1 SEX F\n"
            "1 BIRT\n2 DATE 1850\n",
        )
        rc = run(_args(ged, "surname:Smith sex:M"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "John" in out
        assert "Mary" not in out

    def test_no_matches_returns_success(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME John /Smith/\n",
        )
        rc = run(_args(ged, "surname:Nonexistent"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "No matches" in out

    def test_empty_file_returns_success(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "")
        rc = run(_args(ged, "Smith"))
        assert rc == EXIT_SUCCESS

    def test_no_matches_includes_tip(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "surname:Zymurgy"))
        out = capsys.readouterr().out
        assert "Tip:" in out

    def test_text_output_shows_file_header(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "Smith"))
        out = capsys.readouterr().out
        assert "File:" in out

    def test_text_output_shows_query(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "Smith"))
        out = capsys.readouterr().out
        assert "Query:" in out
        assert "Smith" in out

    def test_sex_filter(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME Alice /Brown/\n1 SEX F\n"
            "0 @I2@ INDI\n1 NAME Bob /Brown/\n1 SEX M\n",
        )
        run(_args(ged, "sex:F"))
        out = capsys.readouterr().out
        assert "Alice" in out
        assert "Bob" not in out

    def test_given_field_search(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME William /Jones/\n"
            "0 @I2@ INDI\n1 NAME Robert /Jones/\n",
        )
        run(_args(ged, "given:William"))
        out = capsys.readouterr().out
        assert "William" in out
        assert "Robert" not in out

    def test_result_count_shown_in_header(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME John /Smith/\n" "0 @I2@ INDI\n1 NAME Jane /Smith/\n",
        )
        run(_args(ged, "surname:Smith"))
        out = capsys.readouterr().out
        assert "2" in out

    def test_wildcard_asterisk_matches(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME John /Smythe/\n"
            "0 @I2@ INDI\n1 NAME Jane /Unrelated/\n",
        )
        run(_args(ged, "surname:Smy*"))
        out = capsys.readouterr().out
        assert "Smythe" in out
        assert "Unrelated" not in out


# ---------------------------------------------------------------------------
# Relationship + field matching
# ---------------------------------------------------------------------------


class TestRelationshipSearch:
    def test_ancestor_finds_descendants(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        rc = run(_args(ged, "ancestor:@I1@"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "James" in out

    def test_ancestor_excludes_root(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        run(_args(ged, "ancestor:@I1@"))
        out = capsys.readouterr().out
        assert "John Smith" not in out

    def test_ancestor_combined_with_surname(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        rc = run(_args(ged, "ancestor:@I1@ surname:Smith"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "James" in out
        # Mary Jones is a descendant but has surname Jones not Smith
        assert "Jones" not in out

    def test_descendant_finds_ancestors(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        rc = run(_args(ged, "descendant:@I3@"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        # I3's parents are I1 and I2
        assert "John" in out or "Mary" in out

    def test_invalid_xref_returns_usage_error(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        rc = run(_args(ged, "ancestor:@I999@"))
        assert rc == EXIT_USAGE_ERROR
        err = capsys.readouterr().err
        assert "@I999@" in err

    def test_invalid_xref_error_has_remediation(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        run(_args(ged, "ancestor:@I999@"))
        err = capsys.readouterr().err
        # Should tell user how to find the right xref
        assert (
            "search" in err.lower()
            or "identifier" in err.lower()
            or "found" in err.lower()
        )

    def test_ancestor_with_date_filter(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        # James (I3) born 1860, within range for ancestor:@I1@
        run(_args(ged, "ancestor:@I1@ born:1850-1900"))
        out = capsys.readouterr().out
        assert "James" in out

    def test_descendant_excludes_root(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        run(_args(ged, "descendant:@I3@"))
        out = capsys.readouterr().out
        assert "James Smith" not in out


# ---------------------------------------------------------------------------
# Mode interactions
# ---------------------------------------------------------------------------


class TestModeInteractions:
    def test_count_text_is_bare_integer(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME John /Smith/\n" "0 @I2@ INDI\n1 NAME Jane /Smith/\n",
        )
        run(_args(ged, "surname:Smith", count=True))
        out = capsys.readouterr().out.strip()
        assert out == "2"

    def test_count_json_outputs_count_key(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME John /Smith/\n" "0 @I2@ INDI\n1 NAME Jane /Smith/\n",
        )
        run(_args(ged, "surname:Smith", count=True, format="json"))
        out = capsys.readouterr().out.strip()
        data = json.loads(out)
        assert data == {"count": 2}

    def test_count_zero_when_no_matches(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "surname:Xyz", count=True))
        out = capsys.readouterr().out.strip()
        assert out == "0"

    def test_count_reports_total_not_limited(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME John /Smith/\n"
            "0 @I2@ INDI\n1 NAME Jane /Smith/\n"
            "0 @I3@ INDI\n1 NAME Jim /Smith/\n",
        )
        # --limit would cap results at 1, but --count should report the full total
        run(_args(ged, "surname:Smith", count=True, limit=1))
        out = capsys.readouterr().out.strip()
        assert out == "3"

    def test_limit_truncates_results(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME Alice /Smith/\n"
            "0 @I2@ INDI\n1 NAME Bob /Smith/\n"
            "0 @I3@ INDI\n1 NAME Carol /Smith/\n",
        )
        run(_args(ged, "surname:Smith", limit=1))
        out = capsys.readouterr().out
        # Exactly one individual entry returned (not three)
        # Truncation notice should appear
        assert "limited to 1" in out

    def test_limit_sets_truncated_flag_in_json(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME Alice /Smith/\n" "0 @I2@ INDI\n1 NAME Bob /Smith/\n",
        )
        run(_args(ged, "surname:Smith", limit=1, format="json"))
        data = json.loads(capsys.readouterr().out)
        assert data["truncated"] is True
        assert len(data["matches"]) == 1

    def test_no_truncation_when_limit_not_reached(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "surname:Smith", limit=10, format="json"))
        data = json.loads(capsys.readouterr().out)
        assert data["truncated"] is False

    def test_quiet_suppresses_headers(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "Smith", quiet=True))
        out = capsys.readouterr().out
        assert "File:" not in out
        assert "Query:" not in out
        assert "===" not in out

    def test_quiet_shows_names(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "Smith", quiet=True))
        out = capsys.readouterr().out
        assert "John Smith" in out

    def test_quiet_no_matches_returns_empty_output(
        self, tmp_path: Path, capsys
    ) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "surname:Xyz", quiet=True))
        out = capsys.readouterr().out.strip()
        assert out == ""

    def test_json_output_is_valid_json(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        rc = run(_args(ged, "Smith", format="json"))
        assert rc == EXIT_SUCCESS
        data = json.loads(capsys.readouterr().out)
        assert isinstance(data, dict)

    def test_json_output_structure(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "Smith", format="json"))
        data = json.loads(capsys.readouterr().out)
        assert "file" in data
        assert "query" in data
        assert "total_individuals" in data
        assert "match_count" in data
        assert "truncated" in data
        assert "matches" in data
        assert isinstance(data["matches"], list)

    def test_json_match_entry_structure(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME John /Smith/\n1 SEX M\n"
            "1 BIRT\n2 DATE 1850\n2 PLAC London\n",
        )
        run(_args(ged, "Smith", format="json"))
        data = json.loads(capsys.readouterr().out)
        assert len(data["matches"]) == 1
        match = data["matches"][0]
        assert "xref" in match
        assert "given_name" in match
        assert "surname" in match
        assert "match_details" in match
        assert isinstance(match["match_details"], list)

    def test_regex_mode_matches(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME John /Smith/\n"
            "0 @I2@ INDI\n1 NAME Jane /Unrelated/\n",
        )
        run(_args(ged, r"name:Sm\w+", regex=True))
        out = capsys.readouterr().out
        assert "Smith" in out
        assert "Unrelated" not in out

    def test_fuzzy_dates_widens_range(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(
            tmp_path,
            "0 @I1@ INDI\n1 NAME Born /Earlier/\n1 BIRT\n2 DATE ABT 1855\n",
        )
        # Without fuzzy, 1850-1853 doesn't include 1855
        run(_args(ged, "born:1850-1853"))
        out_no_fuzzy = capsys.readouterr().out
        # With fuzzy=5, range extends to 1858
        run(_args(ged, "born:1850-1853", fuzzy_dates=5))
        out_fuzzy = capsys.readouterr().out
        assert "Earlier" not in out_no_fuzzy
        assert "Earlier" in out_fuzzy


# ---------------------------------------------------------------------------
# Error message quality
# ---------------------------------------------------------------------------


class TestErrorMessages:
    def test_empty_query_returns_usage_error(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        rc = run(_args(ged, "   "))
        assert rc == EXIT_USAGE_ERROR
        err = capsys.readouterr().err
        assert "Error:" in err

    def test_empty_query_error_has_usage_hint(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, ""))
        err = capsys.readouterr().err
        assert "Usage:" in err or "usage:" in err.lower() or "gedcom-tools" in err

    def test_bad_date_format_returns_usage_error(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        rc = run(_args(ged, "born:notadate"))
        assert rc == EXIT_USAGE_ERROR
        err = capsys.readouterr().err
        assert "Error:" in err

    def test_bad_date_error_has_fix_suggestion(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "born:notadate"))
        err = capsys.readouterr().err
        # Should tell user correct format
        assert "YYYY" in err or "format" in err.lower()

    def test_phonetic_on_date_field_returns_usage_error(
        self, tmp_path: Path, capsys
    ) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        rc = run(_args(ged, "born~1850"))
        assert rc == EXIT_USAGE_ERROR

    def test_phonetic_on_date_error_is_helpful(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "born~1850"))
        err = capsys.readouterr().err
        assert "not supported" in err or "date" in err.lower()

    def test_invalid_xref_format_returns_usage_error(
        self, tmp_path: Path, capsys
    ) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        rc = run(_args(ged, "ancestor:I1"))
        assert rc == EXIT_USAGE_ERROR

    def test_invalid_xref_error_shows_correct_format(
        self, tmp_path: Path, capsys
    ) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        run(_args(ged, "ancestor:I1"))
        err = capsys.readouterr().err
        # Should show the @ delimited form
        assert "@I1@" in err

    def test_file_not_found_returns_error(self, tmp_path: Path, capsys) -> None:
        missing = tmp_path / "does_not_exist.ged"
        rc = run(_args(missing, "Smith"))
        assert rc != EXIT_SUCCESS
        err = capsys.readouterr().err
        assert err.strip()  # some error message present

    def test_unknown_field_returns_usage_error(self, tmp_path: Path, capsys) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        rc = run(_args(ged, "badfield:value"))
        assert rc == EXIT_USAGE_ERROR

    def test_unknown_field_error_lists_valid_fields(
        self, tmp_path: Path, capsys
    ) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "badfield:value"))
        err = capsys.readouterr().err
        # Should list valid fields in the error
        assert "surname" in err or "Valid fields" in err

    def test_reversed_date_range_error_suggests_fix(
        self, tmp_path: Path, capsys
    ) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        run(_args(ged, "born:1900-1800"))
        err = capsys.readouterr().err
        assert "Swap" in err or "swap" in err.lower() or "1800-1900" in err


# ---------------------------------------------------------------------------
# Regression tests with real files
# ---------------------------------------------------------------------------


class TestRegression:
    @pytest.mark.skipif(
        not _SAMPLE_GED.exists(),
        reason="555sample.ged not found",
    )
    def test_sample_search_williams(self, capsys) -> None:
        rc = run(_args(_SAMPLE_GED, "surname:Williams"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "Williams" in out

    @pytest.mark.skipif(
        not _SAMPLE_GED.exists(),
        reason="555sample.ged not found",
    )
    def test_sample_search_no_crash(self, capsys) -> None:
        rc = run(_args(_SAMPLE_GED, "Smith"))
        assert rc == EXIT_SUCCESS

    @pytest.mark.skipif(
        not _SAMPLE_GED.exists(),
        reason="555sample.ged not found",
    )
    def test_sample_json_output_valid(self, capsys) -> None:
        rc = run(_args(_SAMPLE_GED, "Williams", format="json"))
        assert rc == EXIT_SUCCESS
        data = json.loads(capsys.readouterr().out)
        assert "file" in data
        assert "query" in data
        assert "total_individuals" in data
        assert "match_count" in data
        assert "matches" in data

    @pytest.mark.skipif(
        not _SAMPLE_GED.exists(),
        reason="555sample.ged not found",
    )
    def test_sample_total_individuals_positive(self, capsys) -> None:
        run(_args(_SAMPLE_GED, "Williams", format="json"))
        data = json.loads(capsys.readouterr().out)
        assert data["total_individuals"] > 0

    @pytest.mark.skipif(
        not _SAMPLE_GED.exists(),
        reason="555sample.ged not found",
    )
    def test_sample_match_count_consistent(self, capsys) -> None:
        run(_args(_SAMPLE_GED, "surname:Williams", format="json"))
        data = json.loads(capsys.readouterr().out)
        assert data["match_count"] == len(data["matches"])

    @pytest.mark.skipif(
        not _SAMPLE_GED.exists(),
        reason="555sample.ged not found",
    )
    def test_sample_encoding_present_in_json(self, capsys) -> None:
        run(_args(_SAMPLE_GED, "Williams", format="json"))
        data = json.loads(capsys.readouterr().out)
        assert "encoding" in data
        assert "detected" in data["encoding"]

    @pytest.mark.skipif(
        not _ROYAL_GED.exists(),
        reason="royal92.ged not found",
    )
    def test_royal_search_victoria(self, capsys) -> None:
        rc = run(_args(_ROYAL_GED, "given:Victoria"))
        assert rc == EXIT_SUCCESS
        out = capsys.readouterr().out
        assert "Victoria" in out

    @pytest.mark.skipif(
        not _ROYAL_GED.exists(),
        reason="royal92.ged not found",
    )
    def test_royal_json_match_xref_format(self, capsys) -> None:
        run(_args(_ROYAL_GED, "given:Victoria", format="json"))
        data = json.loads(capsys.readouterr().out)
        for match in data["matches"]:
            assert match["xref"].startswith("@")
            assert match["xref"].endswith("@")

    @pytest.mark.skipif(
        not _ROYAL_GED.exists(),
        reason="royal92.ged not found",
    )
    def test_royal_count_mode(self, capsys) -> None:
        run(_args(_ROYAL_GED, "given:Victoria", count=True))
        out = capsys.readouterr().out.strip()
        assert out.isdigit()
        assert int(out) > 0


# ---------------------------------------------------------------------------
# Exit codes
# ---------------------------------------------------------------------------


class TestExitCodes:
    def test_success_with_matches(self, tmp_path: Path) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        assert run(_args(ged, "Smith")) == EXIT_SUCCESS

    def test_success_with_no_matches(self, tmp_path: Path) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        assert run(_args(ged, "surname:Zymurgy")) == EXIT_SUCCESS

    def test_query_parse_error_returns_usage_error(self, tmp_path: Path) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        assert run(_args(ged, "")) == EXIT_USAGE_ERROR

    def test_unknown_field_returns_usage_error(self, tmp_path: Path) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        assert run(_args(ged, "invalid_field:value")) == EXIT_USAGE_ERROR

    def test_file_not_found_returns_nonzero(self, tmp_path: Path) -> None:
        missing = tmp_path / "nowhere.ged"
        rc = run(_args(missing, "Smith"))
        assert rc != EXIT_SUCCESS

    def test_invalid_xref_in_query_returns_usage_error(self, tmp_path: Path) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        # Malformed xref (missing @ delimiters) fails parse_query
        assert run(_args(ged, "ancestor:I1")) == EXIT_USAGE_ERROR

    def test_xref_not_in_file_returns_usage_error(self, tmp_path: Path) -> None:
        ged = _write_ged(tmp_path, _FAMILY_GED)
        assert run(_args(ged, "ancestor:@I999@")) == EXIT_USAGE_ERROR

    def test_via_main_cli_success(self, tmp_path: Path) -> None:
        ged = _write_ged(tmp_path, "0 @I1@ INDI\n1 NAME John /Smith/\n")
        rc = main(["--no-color", "search", str(ged), "Smith"])
        assert rc == EXIT_SUCCESS

    def test_via_main_cli_missing_file(self, tmp_path: Path) -> None:
        missing = tmp_path / "gone.ged"
        rc = main(["--no-color", "search", str(missing), "Smith"])
        assert rc != EXIT_SUCCESS



"""Isolated command — find individuals with no effective family connections."""

from __future__ import annotations

import argparse
import json
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING

from ged4py.parser import GedcomReader

from gedcom_tools.constants import EXIT_ERROR, EXIT_SUCCESS
from gedcom_tools.dates import extract_year_from_date
from gedcom_tools.graph import find_connected_components
from gedcom_tools.progress import Colors, PhaseTracker
from gedcom_tools.utils import extract_xref, validate_input_file

if TYPE_CHECKING:
    from argparse import Namespace, _SubParsersAction


@dataclass
class IsolatedIndividual:

    xref: str
    name: str = ""
    sex: str = ""
    birth_year: int | None = None


@dataclass
class IsolatedResult:

    file_path: str
    total_individuals: int
    singletons: list[IsolatedIndividual] = field(default_factory=list)
    pairs: list[list[IsolatedIndividual]] = field(default_factory=list)

    @property
    def isolated_count(self) -> int:
        return len(self.singletons) + 2 * len(self.pairs)

    def format_text(self, colors: Colors, quiet: bool = False) -> str:
        if quiet:
            if self.isolated_count == 0:
                return ""
            s_label = "singleton" if len(self.singletons) == 1 else "singletons"
            p_label = "pair" if len(self.pairs) == 1 else "pairs"
            return (
                f"{self.isolated_count} isolated "
                f"({len(self.singletons)} {s_label}, "
                f"{len(self.pairs)} {p_label})"
            )

        lines: list[str] = [f"File: {self.file_path}", ""]

        lines.append(_header(colors, "Isolated Analysis"))
        lines.append(f"  Total individuals:    {self.total_individuals:>5}")
        if self.total_individuals > 0:
            pct = self.isolated_count / self.total_individuals * 100
            lines.append(
                f"  Isolated individuals: {self.isolated_count:>5} ({pct:.1f}%)"
            )
        else:
            lines.append(f"  Isolated individuals: {self.isolated_count:>5}")
        lines.append(f"    Singletons:         {len(self.singletons):>5}")
        lines.append(
            f"    Isolated pairs:     {len(self.pairs):>5}"
            + (f" ({len(self.pairs) * 2} individuals)" if self.pairs else "")
        )

        if self.singletons:
            lines.append("")
            lines.append(_header(colors, "Singletons"))
            lines.append("  These individuals have no effective family connections.")
            lines.append(
                "  They may need to be linked to a family or removed if added in error."
            )
            lines.append("")
            for i, ind in enumerate(self.singletons, 1):
                lines.append(f"  {i}. {_format_individual(ind)}")

        if self.pairs:
            lines.append("")
            lines.append(_header(colors, "Isolated Pairs"))
            lines.append(
                "  These pairs are connected only to each other,"
                " with no link to anyone else."
            )
            lines.append("")
            for i, pair in enumerate(self.pairs, 1):
                lines.append(f"  {i}. {_format_individual(pair[0])}")
                if len(pair) > 1:
                    lines.append(f"     {_format_individual(pair[1])}")

        return "\n".join(lines)

    def format_json(self) -> str:
        data = {
            "file": self.file_path,
            "summary": {
                "total_individuals": self.total_individuals,
                "isolated_count": self.isolated_count,
                "singleton_count": len(self.singletons),
                "pair_count": len(self.pairs),
            },
            "singletons": [_individual_to_dict(s) for s in self.singletons],
            "pairs": [[_individual_to_dict(p) for p in pair] for pair in self.pairs],
        }
        return json.dumps(data, indent=2)


def _header(colors: Colors, title: str) -> str:
    return f"{colors.cyan}=== {title} ==={colors.reset}"


def _format_individual(ind: IsolatedIndividual) -> str:
    parts = [f"{ind.name} ({ind.xref})"]
    if ind.sex:
        parts.append(ind.sex)
    if ind.birth_year is not None:
        parts.append(f"b. {ind.birth_year}")
    if len(parts) == 1:
        return parts[0]
    return f"{parts[0]} {', '.join(parts[1:])}"


def _individual_to_dict(ind: IsolatedIndividual) -> dict[str, str | int | None]:
    return {
        "xref": ind.xref,
        "name": ind.name,
        "sex": ind.sex,
        "birth_year": ind.birth_year,
    }


_extract_xref = extract_xref


def _collect_data(
    file_path: Path,
    *,
    quiet: bool,
    verbose: bool,
    no_color: bool,
) -> IsolatedResult:
    """Collect individuals and families, then find isolated."""
    tracker = PhaseTracker(
        total_phases=2,
        stream=sys.stderr,
        no_color=no_color,
        quiet=quiet,
        verbose=verbose,
    )

    individuals: dict[str, IsolatedIndividual] = {}
    family_members: dict[str, list[str]] = {}

    # Phase 1: Collect data
    with tracker.phase("Collecting data"):
        with GedcomReader(str(file_path)) as reader:
            for rec in reader.records0("INDI"):
                xref = rec.xref_id
                if not xref:
                    continue

                name = ""
                name_rec = rec.sub_tag("NAME")
                if name_rec and name_rec.value:
                    val = name_rec.value
                    if isinstance(val, tuple):
                        # ged4py returns (given, surname, suffix)
                        name = " ".join(p for p in val if p).strip()
                    else:
                        name = str(val).replace("/", "").strip()

                sex = ""
                sex_rec = rec.sub_tag("SEX")
                if sex_rec and sex_rec.value:
                    sex = str(sex_rec.value).strip()

                birth_year = None
                birt_rec = rec.sub_tag("BIRT")
                if birt_rec:
                    date_rec = birt_rec.sub_tag("DATE")
                    if date_rec and date_rec.value:
                        birth_year = extract_year_from_date(date_rec.value)

                individuals[xref] = IsolatedIndividual(
                    xref=xref, name=name, sex=sex, birth_year=birth_year
                )

            for rec in reader.records0("FAM"):
                xref = rec.xref_id
                if not xref:
                    continue

                members: list[str] = []
                for sub in rec.sub_records:
                    if sub.tag in ("HUSB", "WIFE", "CHIL") and sub.value:
                        m = _extract_xref(sub.value)
                        if m:
                            members.append(m)

                family_members[xref] = members

    # Phase 2: Find isolated
    with tracker.phase("Analyzing connections"):
        components = find_connected_components(set(individuals.keys()), family_members)

        singletons: list[IsolatedIndividual] = []
        pairs: list[list[IsolatedIndividual]] = []

        for members in components.values():
            if len(members) == 1:
                singletons.append(individuals[members[0]])
            elif len(members) == 2:
                pairs.append([individuals[members[0]], individuals[members[1]]])

    return IsolatedResult(
        file_path=str(file_path),
        total_individuals=len(individuals),
        singletons=singletons,
        pairs=pairs,
    )


def register_subcommand(
    subparsers: _SubParsersAction[argparse.ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "isolated",
        help="Find isolated individuals with no family connections",
        description=(
            "Detect individuals in connected components of size 1 (singletons)"
            " or size 2 (isolated pairs) using graph analysis."
        ),
    )
    parser.add_argument(
        "file",
        type=Path,
        help="Path to the GEDCOM file to analyze",
    )


def run(args: Namespace) -> int:
    file_path: Path = args.file

    if err := validate_input_file(file_path):
        return err

    output_format = getattr(args, "format", "text")
    quiet = getattr(args, "quiet", False)
    verbose = getattr(args, "verbose", False)
    no_color = getattr(args, "no_color", False)

    try:
        result = _collect_data(
            file_path, quiet=quiet, verbose=verbose, no_color=no_color
        )

        if output_format == "json":
            print(result.format_json())
        else:
            colors = Colors(sys.stdout, force_disable=no_color)
            output = result.format_text(colors, quiet=quiet)
            if output:
                print(output)

        return EXIT_SUCCESS

    except Exception as e:
        if verbose:
            raise
        print(f"Error: {e}", file=sys.stderr)
        return EXIT_ERROR



from __future__ import annotations

import argparse
import json
import unicodedata
from pathlib import Path

import ansel  # type: ignore[import-untyped]
import pytest

from gedcom_tools.commands.convert import run
from gedcom_tools.commands.convert.transcoder import (
    CODEC_TO_CHAR,
    ConvertResult,
    count_long_lines,
    resolve_target_codec,
    transcode,
    update_char_header,
)
from gedcom_tools.constants import EXIT_ERROR, EXIT_SUCCESS, EXIT_USAGE_ERROR
from gedcom_tools.progress import Colors
from gedcom_tools.utils import BOMS, EncodingInfo, resolve_source_codec, strip_bom

ansel.register()

MINIMAL_GED = (
    "0 HEAD\n1 SOUR TEST\n1 GEDC\n2 VERS 5.5.1\n1 CHAR UTF-8\n"
    "0 @I1@ INDI\n1 NAME John /Smith/\n0 TRLR\n"
)


def _write_ged(path: Path, content: str = MINIMAL_GED) -> Path:
    ged = path / "test.ged"
    ged.write_text(content, encoding="utf-8")
    return ged


def _make_args(
    file_path: Path,
    to_encoding: str,
    output: Path,
    **kwargs: object,
) -> argparse.Namespace:
    defaults: dict[str, object] = {
        "file": file_path,
        "to_encoding": to_encoding,
        "output": output,
        "from_encoding": None,
        "force": False,
        "bom": False,
        "no_normalize": False,
        "dry_run": False,
        "verbose": False,
        "quiet": False,
        "no_color": True,
        "format": "text",
    }
    defaults.update(kwargs)
    return argparse.Namespace(**defaults)


# ---------------------------------------------------------------------------
# TestResolveSourceCodec
# ---------------------------------------------------------------------------


class TestResolveSourceCodec:
    def test_ansel_from_detect(self) -> None:
        info = EncodingInfo(encoding="ANSEL")
        assert resolve_source_codec(info, None) == "gedcom"

    def test_utf8_from_detect(self) -> None:
        info = EncodingInfo(encoding="UTF-8")
        assert resolve_source_codec(info, None) == "utf-8"

    def test_utf16le_from_detect(self) -> None:
        info = EncodingInfo(encoding="UTF-16-LE")
        assert resolve_source_codec(info, None) == "utf-16-le"

    def test_utf16be_from_detect(self) -> None:
        info = EncodingInfo(encoding="UTF-16-BE")
        assert resolve_source_codec(info, None) == "utf-16-be"

    def test_override_gedcom_name(self) -> None:
        info = EncodingInfo(encoding="UTF-8")
        assert resolve_source_codec(info, "ansel") == "gedcom"

    def test_override_python_name(self) -> None:
        info = EncodingInfo(encoding="UTF-8")
        result = resolve_source_codec(info, "latin-1")
        assert result == "iso8859-1"

    def test_override_invalid(self) -> None:
        info = EncodingInfo(encoding="UTF-8")
        with pytest.raises(ValueError, match="Unknown source encoding"):
            resolve_source_codec(info, "not-a-codec")

    def test_ascii_from_detect(self) -> None:
        info = EncodingInfo(encoding="ASCII")
        assert resolve_source_codec(info, None) == "ascii"

    def test_unicode_from_detect(self) -> None:
        info = EncodingInfo(encoding="UNICODE")
        assert resolve_source_codec(info, None) == "utf-16-le"

    def test_case_insensitive_lookup(self) -> None:
        # "Utf-8" is not in SOURCE_ENCODING_MAP, but "utf-8" is in GEDCOM_CHARSETS
        info = EncodingInfo(encoding="Utf-8")
        assert resolve_source_codec(info, None) == "utf-8"

    def test_detect_fallback_codecs_lookup(self) -> None:
        # "iso-8859-1" is not in SOURCE_ENCODING_MAP or GEDCOM_CHARSETS,
        # but codecs.lookup("iso-8859-1") succeeds.
        info = EncodingInfo(encoding="iso-8859-1")
        assert resolve_source_codec(info, None) == "iso8859-1"

    def test_detect_unknown_encoding_raises(self) -> None:
        info = EncodingInfo(encoding="BOGUS-ENCODING")
        with pytest.raises(ValueError, match="Cannot determine source encoding"):
            resolve_source_codec(info, None)


# ---------------------------------------------------------------------------
# TestResolveTargetCodec
# ---------------------------------------------------------------------------


class TestResolveTargetCodec:
    def test_utf8(self) -> None:
        assert resolve_target_codec("utf-8") == "utf-8"

    def test_ansel(self) -> None:
        assert resolve_target_codec("ansel") == "gedcom"

    def test_ascii(self) -> None:
        assert resolve_target_codec("ascii") == "ascii"

    def test_unicode(self) -> None:
        assert resolve_target_codec("unicode") == "utf-16-le"


# ---------------------------------------------------------------------------
# TestStripBom
# ---------------------------------------------------------------------------


class TestStripBom:
    def test_utf8_bom(self) -> None:
        data = b"\xef\xbb\xbfHello"
        stripped, enc = strip_bom(data)
        assert stripped == b"Hello"
        assert enc == "utf-8"

    def test_utf16le_bom(self) -> None:
        data = b"\xff\xfeH\x00e\x00"
        stripped, enc = strip_bom(data)
        assert stripped == b"H\x00e\x00"
        assert enc == "utf-16-le"

    def test_utf16be_bom(self) -> None:
        data = b"\xfe\xff\x00H\x00e"
        stripped, enc = strip_bom(data)
        assert stripped == b"\x00H\x00e"
        assert enc == "utf-16-be"

    def test_no_bom(self) -> None:
        data = b"plain text"
        stripped, enc = strip_bom(data)
        assert stripped == data
        assert enc is None

    def test_empty(self) -> None:
        stripped, enc = strip_bom(b"")
        assert stripped == b""
        assert enc is None

    def test_utf8_bom_priority(self) -> None:
        # UTF-8 BOM (3 bytes) followed by bytes that look like UTF-16-LE BOM
        data = b"\xef\xbb\xbf\xff\xfe"
        stripped, enc = strip_bom(data)
        assert stripped == b"\xff\xfe"
        assert enc == "utf-8"

    def test_returns_bom_encoding(self) -> None:
        for bom_name, bom_bytes in BOMS.items():
            _, enc = strip_bom(bom_bytes + b"data")
            assert enc == bom_name


# ---------------------------------------------------------------------------
# TestUpdateCharHeader
# ---------------------------------------------------------------------------


class TestUpdateCharHeader:
    def test_replace_existing(self) -> None:
        text = "0 HEAD\n1 CHAR UTF-8\n0 TRLR\n"
        result = update_char_header(text, "ANSEL")
        assert "1 CHAR ANSEL" in result
        assert "1 CHAR UTF-8" not in result

    def test_replace_empty_char(self) -> None:
        text = "0 HEAD\n1 CHAR\n0 TRLR\n"
        result = update_char_header(text, "UTF-8")
        assert "1 CHAR UTF-8" in result

    def test_insert_when_missing_with_subrecords(self) -> None:
        text = "0 HEAD\n1 SOUR TEST\n1 GEDC\n2 VERS 5.5.1\n0 @I1@ INDI\n0 TRLR\n"
        result = update_char_header(text, "UTF-8")
        assert "1 CHAR UTF-8" in result
        # CHAR should appear after HEAD line
        lines = result.split("\n")
        head_idx = next(i for i, ln in enumerate(lines) if ln.startswith("0 HEAD"))
        char_idx = next(i for i, ln in enumerate(lines) if "1 CHAR UTF-8" in ln)
        assert char_idx == head_idx + 1

    def test_insert_when_missing_no_subrecords(self) -> None:
        text = "0 HEAD\n0 @I1@ INDI\n0 TRLR\n"
        result = update_char_header(text, "UTF-8")
        assert "1 CHAR UTF-8" in result

    def test_preserves_other_lines(self) -> None:
        text = "0 HEAD\n1 SOUR TEST\n1 CHAR UTF-8\n0 @I1@ INDI\n0 TRLR\n"
        result = update_char_header(text, "ANSEL")
        assert "1 SOUR TEST" in result
        assert "0 @I1@ INDI" in result
        assert "0 TRLR" in result

    def test_crlf_line_endings(self) -> None:
        text = "0 HEAD\r\n1 CHAR UTF-8\r\n0 TRLR\r\n"
        result = update_char_header(text, "ANSEL")
        assert "1 CHAR ANSEL\r\n" in result

    def test_crlf_insert_uses_crlf(self) -> None:
        text = "0 HEAD\r\n1 SOUR TEST\r\n0 TRLR\r\n"
        result = update_char_header(text, "UTF-8")
        assert "1 CHAR UTF-8\r\n" in result

    def test_no_head_raises_error(self) -> None:
        text = "1 SOUR TEST\n0 TRLR\n"
        with pytest.raises(ValueError, match="No HEAD record found"):
            update_char_header(text, "UTF-8")

    def test_does_not_match_charset(self) -> None:
        # "1 CHARSET ANSI" should NOT be matched by the CHAR regex
        text = "0 HEAD\n1 CHARSET ANSI\n1 CHAR UTF-8\n0 TRLR\n"
        result = update_char_header(text, "ANSEL")
        assert "1 CHARSET ANSI" in result
        assert "1 CHAR ANSEL" in result


# ---------------------------------------------------------------------------
# TestCountLongLines
# ---------------------------------------------------------------------------


class TestCountLongLines:
    def test_all_short(self) -> None:
        text = "short line\nanother\nthird\n"
        total, over = count_long_lines(text, "utf-8")
        assert total == 3
        assert over == 0

    def test_some_long(self) -> None:
        short = "ok\n"
        long_line = "x" * 256 + "\n"
        text = short + long_line + short
        total, over = count_long_lines(text, "utf-8")
        assert total == 3
        assert over == 1

    def test_utf8_multibyte(self) -> None:
        # Each "\u00e9" is 2 bytes in UTF-8. 128 of them = 256 bytes > 255 limit.
        line = "\u00e9" * 128 + "\n"
        total, over = count_long_lines(line, "utf-8")
        assert total == 1
        assert over == 1

    def test_empty(self) -> None:
        total, over = count_long_lines("", "utf-8")
        assert total == 0
        assert over == 0

    def test_utf16le_byte_lengths(self) -> None:
        # In UTF-16-LE, each ASCII char is 2 bytes. 128 chars = 256 bytes > 255.
        line = "a" * 128 + "\n"
        total, over = count_long_lines(line, "utf-16-le")
        assert total == 1
        assert over == 1


# ---------------------------------------------------------------------------
# TestTranscode
# ---------------------------------------------------------------------------


class TestTranscode:
    def test_utf8_to_utf8(self, tmp_path: Path) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        result = transcode(
            src,
            out,
            source_codec="utf-8",
            target_codec="utf-8",
            target_char="UTF-8",
            normalize=False,
            add_bom=False,
            dry_run=False,
        )
        assert out.exists()
        content = out.read_text(encoding="utf-8")
        assert "John /Smith/" in content
        assert result.lines_total > 0

    def test_utf8_to_utf16(self, tmp_path: Path) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        result = transcode(
            src,
            out,
            source_codec="utf-8",
            target_codec="utf-16-le",
            target_char="UNICODE",
            normalize=False,
            add_bom=False,
            dry_run=False,
        )
        raw = out.read_bytes()
        text = raw.decode("utf-16-le")
        assert "1 CHAR UNICODE" in text
        assert result.target_encoding == "UNICODE"

    def test_utf16_source_to_utf8(self, tmp_path: Path) -> None:
        src = tmp_path / "src.ged"
        bom = b"\xff\xfe"
        encoded = MINIMAL_GED.encode("utf-16-le")
        src.write_bytes(bom + encoded)
        out = tmp_path / "out.ged"
        result = transcode(
            src,
            out,
            source_codec="utf-16-le",
            target_codec="utf-8",
            target_char="UTF-8",
            normalize=False,
            add_bom=False,
            dry_run=False,
        )
        content = out.read_text(encoding="utf-8")
        assert "1 CHAR UTF-8" in content
        assert result.bom_stripped == "utf-16-le"

    def test_nfc_normalization(self, tmp_path: Path) -> None:
        # NFD: e followed by combining acute accent
        nfd_name = "e\u0301"
        ged_text = MINIMAL_GED.replace("John", nfd_name)
        src = tmp_path / "src.ged"
        src.write_text(ged_text, encoding="utf-8")
        out = tmp_path / "out.ged"
        transcode(
            src,
            out,
            source_codec="utf-8",
            target_codec="utf-8",
            target_char="UTF-8",
            normalize=True,
            add_bom=False,
            dry_run=False,
        )
        content = out.read_text(encoding="utf-8")
        assert "\u00e9" in content
        # The NFD form should be gone
        assert "e\u0301" not in content

    def test_no_normalize(self, tmp_path: Path) -> None:
        nfd_name = "e\u0301"
        ged_text = MINIMAL_GED.replace("John", nfd_name)
        src = tmp_path / "src.ged"
        src.write_text(ged_text, encoding="utf-8")
        out = tmp_path / "out.ged"
        transcode(
            src,
            out,
            source_codec="utf-8",
            target_codec="utf-8",
            target_char="UTF-8",
            normalize=False,
            add_bom=False,
            dry_run=False,
        )
        content = out.read_text(encoding="utf-8")
        # NFD should be preserved when normalize=False
        nfc = unicodedata.normalize("NFC", content)
        assert content != nfc  # content still has NFD sequences

    def test_bom_stripped(self, tmp_path: Path) -> None:
        src = tmp_path / "src.ged"
        src.write_bytes(b"\xef\xbb\xbf" + MINIMAL_GED.encode("utf-8"))
        out = tmp_path / "out.ged"
        result = transcode(
            src,
            out,
            source_codec="utf-8",
            target_codec="utf-8",
            target_char="UTF-8",
            normalize=False,
            add_bom=False,
            dry_run=False,
        )
        assert result.bom_stripped == "utf-8"

    def test_bom_added(self, tmp_path: Path) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        transcode(
            src,
            out,
            source_codec="utf-8",
            target_codec="utf-8",
            target_char="UTF-8",
            normalize=False,
            add_bom=True,
            dry_run=False,
        )
        raw = out.read_bytes()
        assert raw[:3] == b"\xef\xbb\xbf"

    def test_bom_ignored_ascii(self, tmp_path: Path) -> None:
        # ASCII is not in BOM_ENCODINGS, so run() would set add_bom=False.
        # Calling transcode with add_bom=False for ascii target: no BOM.
        ascii_ged = "0 HEAD\n1 SOUR TEST\n1 CHAR ASCII\n0 TRLR\n"
        src = tmp_path / "src.ged"
        src.write_text(ascii_ged, encoding="ascii")
        out = tmp_path / "out.ged"
        transcode(
            src,
            out,
            source_codec="ascii",
            target_codec="ascii",
            target_char="ASCII",
            normalize=False,
            add_bom=False,
            dry_run=False,
        )
        raw = out.read_bytes()
        # No BOM prefix
        assert raw[:3] != b"\xef\xbb\xbf"
        assert raw[:2] != b"\xff\xfe"
        assert raw[:2] != b"\xfe\xff"

    def test_char_updated(self, tmp_path: Path) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        transcode(
            src,
            out,
            source_codec="utf-8",
            target_codec="utf-16-le",
            target_char="UNICODE",
            normalize=False,
            add_bom=False,
            dry_run=False,
        )
        content = out.read_bytes().decode("utf-16-le")
        assert "1 CHAR UNICODE" in content
        assert "1 CHAR UTF-8" not in content

    def test_dry_run_no_write(self, tmp_path: Path) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        result = transcode(
            src,
            out,
            source_codec="utf-8",
            target_codec="utf-8",
            target_char="UTF-8",
            normalize=False,
            add_bom=False,
            dry_run=True,
        )
        assert not out.exists()
        assert result.dry_run is True
        assert result.lines_total > 0

    def test_decode_error(self, tmp_path: Path) -> None:
        src = tmp_path / "bad.ged"
        # 0x80 is not valid UTF-8 start byte
        src.write_bytes(b"0 HEAD\n1 CHAR UTF-8\n\x80\n0 TRLR\n")
        out = tmp_path / "out.ged"
        with pytest.raises(ValueError, match="Failed to decode"):
            transcode(
                src,
                out,
                source_codec="utf-8",
                target_codec="utf-8",
                target_char="UTF-8",
                normalize=False,
                add_bom=False,
                dry_run=False,
            )

    def test_encode_error_utf8_to_ascii(self, tmp_path: Path) -> None:
        ged_text = MINIMAL_GED.replace("John", "\u00e9mile")
        src = tmp_path / "src.ged"
        src.write_text(ged_text, encoding="utf-8")
        out = tmp_path / "out.ged"
        with pytest.raises(ValueError, match=r"U\+00E9"):
            transcode(
                src,
                out,
                source_codec="utf-8",
                target_codec="ascii",
                target_char="ASCII",
                normalize=False,
                add_bom=False,
                dry_run=False,
            )

    def test_line_endings_crlf_preserved(self, tmp_path: Path) -> None:
        crlf_ged = MINIMAL_GED.replace("\n", "\r\n")
        src = tmp_path / "src.ged"
        src.write_text(crlf_ged, encoding="utf-8", newline="")
        out = tmp_path / "out.ged"
        transcode(
            src,
            out,
            source_codec="utf-8",
            target_codec="utf-8",
            target_char="UTF-8",
            normalize=False,
            add_bom=False,
            dry_run=False,
        )
        raw = out.read_bytes()
        assert b"\r\n" in raw

    def test_empty_file_raises(self, tmp_path: Path) -> None:
        src = tmp_path / "empty.ged"
        src.write_bytes(b"")
        out = tmp_path / "out.ged"
        with pytest.raises(ValueError, match="File is empty"):
            transcode(
                src,
                out,
                source_codec="utf-8",
                target_codec="utf-8",
                target_char="UTF-8",
                normalize=False,
                add_bom=False,
                dry_run=False,
            )

    def test_bom_stripped_value(self, tmp_path: Path) -> None:
        src = tmp_path / "bom.ged"
        src.write_bytes(BOMS["utf-16-le"] + MINIMAL_GED.encode("utf-16-le"))
        out = tmp_path / "out.ged"
        result = transcode(
            src,
            out,
            source_codec="utf-16-le",
            target_codec="utf-8",
            target_char="UTF-8",
            normalize=False,
            add_bom=False,
            dry_run=False,
        )
        assert result.bom_stripped == "utf-16-le"

    def test_source_encoding_label(self, tmp_path: Path) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        result = transcode(
            src,
            out,
            source_codec="utf-8",
            target_codec="utf-8",
            target_char="UTF-8",
            normalize=False,
            add_bom=False,
            dry_run=False,
        )
        assert result.source_encoding == CODEC_TO_CHAR["utf-8"]
        assert result.source_encoding == "UTF-8"


# ---------------------------------------------------------------------------
# TestConvertResultFormatText
# ---------------------------------------------------------------------------


class TestConvertResultFormatText:
    def _make_result(self, **kwargs: object) -> ConvertResult:
        defaults: dict[str, object] = {
            "source_file": Path("input.ged"),
            "output_file": Path("output.ged"),
            "source_encoding": "UTF-8",
            "target_encoding": "UTF-8",
            "source_codec": "utf-8",
            "target_codec": "utf-8",
            "lines_total": 10,
            "lines_over_limit": 0,
            "normalized": False,
            "bom_added": False,
            "bom_stripped": None,
            "dry_run": False,
        }
        defaults.update(kwargs)
        return ConvertResult(**defaults)

    def test_quiet_dry_run(self) -> None:
        result = self._make_result(dry_run=True)
        colors = Colors(force_disable=True)
        text = result.format_text(colors, quiet=True)
        assert "(dry run)" in text
        assert "Converted" in text

    def test_bom_stripped_label(self) -> None:
        result = self._make_result(bom_stripped="utf-8")
        colors = Colors(force_disable=True)
        text = result.format_text(colors, quiet=False)
        assert "stripped" in text

    def test_bom_added_label(self) -> None:
        result = self._make_result(bom_added=True)
        colors = Colors(force_disable=True)
        text = result.format_text(colors, quiet=False)
        assert "added" in text

    def test_lines_over_limit_warning(self) -> None:
        result = self._make_result(lines_over_limit=5)
        colors = Colors(force_disable=True)
        text = result.format_text(colors, quiet=False)
        assert "5 lines" in text
        assert "255-byte" in text

    def test_dry_run_verbose(self) -> None:
        result = self._make_result(dry_run=True)
        colors = Colors(force_disable=True)
        text = result.format_text(colors, quiet=False)
        assert "dry run" in text
        assert "no file written" in text

    def test_format_json(self) -> None:
        result = self._make_result()
        output = json.loads(result.format_json())
        assert output["source_encoding"] == "UTF-8"
        assert output["target_encoding"] == "UTF-8"
        assert "gedcom_tools_version" in output


# ---------------------------------------------------------------------------
# TestConvertRun
# ---------------------------------------------------------------------------


class TestConvertRun:
    def test_basic_roundtrip(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        code = run(_make_args(src, "utf-8", out))
        assert code == EXIT_SUCCESS
        content = out.read_text(encoding="utf-8")
        assert "John /Smith/" in content

    def test_overwrite_protection(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        out.write_text("existing", encoding="utf-8")
        code = run(_make_args(src, "utf-8", out))
        assert code == EXIT_ERROR
        captured = capsys.readouterr()
        assert "already exists" in captured.err

    def test_force_overwrite(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        out.write_text("existing", encoding="utf-8")
        code = run(_make_args(src, "utf-8", out, force=True))
        assert code == EXIT_SUCCESS
        content = out.read_text(encoding="utf-8")
        assert "John /Smith/" in content

    def test_same_file_rejected(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        src = _write_ged(tmp_path)
        # Use force=True to bypass overwrite protection (output exists = input)
        code = run(_make_args(src, "utf-8", src, force=True))
        assert code == EXIT_ERROR
        captured = capsys.readouterr()
        assert "resolves to the input" in captured.err

    def test_same_file_via_symlink(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        src = _write_ged(tmp_path)
        link = tmp_path / "link.ged"
        link.symlink_to(src)
        # Use force=True to bypass overwrite protection (symlink exists)
        code = run(_make_args(src, "utf-8", link, force=True))
        assert code == EXIT_ERROR
        captured = capsys.readouterr()
        assert "resolves to the input" in captured.err

    def test_missing_input(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        missing = tmp_path / "nonexistent.ged"
        out = tmp_path / "out.ged"
        code = run(_make_args(missing, "utf-8", out))
        assert code == EXIT_USAGE_ERROR

    def test_invalid_source(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        # Write valid GEDCOM with latin-1 high bytes, then try to decode as UTF-8.
        # detect_encoding sees "1 CHAR ASCII" and ged4py parses fine,
        # but --from utf-8 override makes transcode fail on the high byte.
        latin1_ged = (
            "0 HEAD\n1 SOUR TEST\n1 CHAR ASCII\n"
            "0 @I1@ INDI\n1 NAME Ren\xe9 /Test/\n0 TRLR\n"
        )
        src = tmp_path / "bad.ged"
        src.write_bytes(latin1_ged.encode("latin-1"))
        out = tmp_path / "out.ged"
        code = run(_make_args(src, "utf-8", out, from_encoding="utf-8"))
        assert code == EXIT_ERROR

    def test_dry_run(self, tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        code = run(_make_args(src, "utf-8", out, dry_run=True))
        assert code == EXIT_SUCCESS
        assert not out.exists()
        captured = capsys.readouterr()
        assert "dry run" in captured.out

    def test_quiet_mode(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        code = run(_make_args(src, "utf-8", out, quiet=True))
        assert code == EXIT_SUCCESS
        captured = capsys.readouterr()
        output_text = captured.out.strip()
        # Quiet mode produces a single-line summary
        assert output_text.count("\n") == 0
        assert "Converted" in output_text

    def test_json_format(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        code = run(_make_args(src, "utf-8", out, format="json"))
        assert code == EXIT_SUCCESS
        captured = capsys.readouterr()
        data = json.loads(captured.out)
        assert data["source_encoding"] == "UTF-8"
        assert data["target_encoding"] == "UTF-8"
        assert "gedcom_tools_version" in data

    def test_bom_flag(self, tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        code = run(_make_args(src, "utf-8", out, bom=True))
        assert code == EXIT_SUCCESS
        raw = out.read_bytes()
        assert raw[:3] == b"\xef\xbb\xbf"

    def test_no_normalize_flag(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        # With ANSEL source, normalize would normally be True.
        # Use --no-normalize to skip it. We'll test via --from override.
        nfd_name = "e\u0301"
        ged_text = MINIMAL_GED.replace("John", nfd_name)
        src = tmp_path / "src.ged"
        src.write_text(ged_text, encoding="utf-8")
        out = tmp_path / "out.ged"
        code = run(
            _make_args(
                src,
                "utf-8",
                out,
                from_encoding="utf-8",
                no_normalize=True,
            )
        )
        assert code == EXIT_SUCCESS
        content = out.read_text(encoding="utf-8")
        # NFD sequence should be preserved
        assert "e\u0301" in content

    def test_long_line_warning(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        long_line = "1 NOTE " + "x" * 260
        ged_text = MINIMAL_GED.replace("1 NAME John /Smith/", long_line)
        src = tmp_path / "src.ged"
        src.write_text(ged_text, encoding="utf-8")
        out = tmp_path / "out.ged"
        code = run(_make_args(src, "utf-8", out))
        assert code == EXIT_SUCCESS
        captured = capsys.readouterr()
        assert "255-byte" in captured.err or "exceed" in captured.err

    def test_from_override_latin1(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        # Write a latin-1 encoded file
        latin1_text = (
            "0 HEAD\n1 SOUR TEST\n1 CHAR ASCII\n"
            "0 @I1@ INDI\n1 NAME \xe9mile /Duval/\n0 TRLR\n"
        )
        src = tmp_path / "src.ged"
        src.write_bytes(latin1_text.encode("latin-1"))
        out = tmp_path / "out.ged"
        code = run(_make_args(src, "utf-8", out, from_encoding="latin-1"))
        assert code == EXIT_SUCCESS
        content = out.read_text(encoding="utf-8")
        assert "\u00e9mile" in content

    def test_invalid_from_encoding(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        src = _write_ged(tmp_path)
        out = tmp_path / "out.ged"
        code = run(_make_args(src, "utf-8", out, from_encoding="not-a-codec"))
        assert code == EXIT_USAGE_ERROR
        captured = capsys.readouterr()
        assert "Unknown source encoding" in captured.err

    def test_from_override_ansel(
        self, tmp_path: Path, capsys: pytest.CaptureFixture[str]
    ) -> None:
        # Build a minimal ANSEL-encoded file with high-byte ANSEL sequences.
        # ANSEL: combining acute (0xE2) precedes base char 'e' (0x65) -> e-acute
        ansel_bytes = (
            b"0 HEAD\r\n"
            b"1 SOUR TEST\r\n"
            b"1 CHAR ANSEL\r\n"
            b"0 @I1@ INDI\r\n"
            b"1 NAME \xe2" + b"emile /Duval/\r\n"
            b"0 TRLR\r\n"
        )
        src = tmp_path / "src.ged"
        src.write_bytes(ansel_bytes)
        out = tmp_path / "out.ged"
        code = run(_make_args(src, "utf-8", out, from_encoding="ansel"))
        assert code == EXIT_SUCCESS
        content = out.read_text(encoding="utf-8")
        # ANSEL codec decodes \xe2 + e as e + combining acute (NFD).
        # run() normalizes ANSEL sources to NFC by default.
        assert "\u00e9" in content or "e\u0301" in content



"""Export GEDCOM data to CSV or JSON."""

from __future__ import annotations

import argparse
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from gedcom_tools.commands.export.collector import collect_export_data
from gedcom_tools.commands.export.formatters import format_csv, format_json
from gedcom_tools.constants import EXIT_ERROR, EXIT_SUCCESS, EXIT_USAGE_ERROR
from gedcom_tools.progress import PhaseTracker
from gedcom_tools.utils import validate_input_file

if TYPE_CHECKING:
    from argparse import Namespace, _SubParsersAction


def register_subcommand(subparsers: _SubParsersAction[argparse.ArgumentParser]) -> None:
    parser = subparsers.add_parser(
        "export",
        help="Export individuals and families to CSV or JSON",
        description=(
            "Extract all individuals and families from a GEDCOM file "
            "into CSV or JSON format for use in spreadsheets, databases, "
            "and downstream tools."
        ),
    )
    parser.add_argument(
        "file",
        type=Path,
        help="GEDCOM file to export",
    )
    parser.add_argument(
        "--format",
        choices=["csv", "json"],
        default=argparse.SUPPRESS,
        help="Export format (default: csv)",
    )
    parser.add_argument(
        "--table",
        choices=["individuals", "families"],
        default="individuals",
        help="Table to export in CSV mode (default: individuals; ignored for JSON)",
    )
    parser.add_argument(
        "--no-bom",
        action="store_true",
        help="Omit UTF-8 BOM when writing CSV to a file",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=None,
        help="Write output to file instead of stdout",
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="Overwrite output file if it already exists",
    )
    parser.add_argument(
        "--redact-living",
        action="store_true",
        help="Replace names and dates of estimated-living individuals",
    )
    parser.add_argument(
        "--max-age",
        type=int,
        default=110,
        help="Maximum age for living estimation (default: 110)",
    )


def run(args: Namespace) -> int:
    file_path: Path = args.file
    quiet: bool = getattr(args, "quiet", False)
    verbose: bool = getattr(args, "verbose", False)
    no_color: bool = getattr(args, "no_color", False)
    table: str = args.table
    redact_living: bool = args.redact_living
    max_age: int = args.max_age
    no_bom: bool = args.no_bom
    output_path: Path | None = getattr(args, "output", None)
    force: bool = getattr(args, "force", False)

    # Resolve export format
    fmt = getattr(args, "format", "text")
    if fmt == "text":
        fmt = "csv"
    elif fmt not in ("csv", "json"):
        print(
            f"Error: --format {fmt} is not valid for the export command. "
            "Use --format csv or --format json.",
            file=sys.stderr,
        )
        return EXIT_USAGE_ERROR

    if err := validate_input_file(file_path):
        return err

    # Overwrite protection
    if output_path and output_path.exists() and not force:
        print(
            f"Error: {output_path} already exists. Use --force to overwrite.",
            file=sys.stderr,
        )
        return EXIT_ERROR

    try:
        tracker = PhaseTracker(
            2, stream=sys.stderr, no_color=no_color, quiet=quiet, verbose=verbose
        )

        with tracker.phase("Reading file"):
            result = collect_export_data(file_path)

        with tracker.phase("Formatting output"):
            include_bom = fmt == "csv" and output_path is not None and not no_bom

            if fmt == "json":
                output = format_json(
                    result, redact_living=redact_living, max_age=max_age
                )
            else:
                output = format_csv(
                    result,
                    table=table,
                    include_bom=include_bom,
                    redact_living=redact_living,
                    max_age=max_age,
                )

        if output_path:
            output_path.write_text(output, encoding="utf-8")
            if not quiet:
                print(f"Exported to {output_path}", file=sys.stderr)
        else:
            sys.stdout.write(output)

        return EXIT_SUCCESS

    except Exception as e:
        if verbose:
            raise
        print(f"Error: {e}", file=sys.stderr)
        return EXIT_ERROR



"""CSV and JSON formatters for the export command."""

from __future__ import annotations

import csv
import io
import json
from typing import TYPE_CHECKING, Any

from gedcom_tools import __version__
from gedcom_tools.commands.export.models import estimate_living

if TYPE_CHECKING:
    from gedcom_tools.commands.export.models import (
        ExportFamily,
        ExportIndividual,
        ExportResult,
    )

_INDI_CSV_COLUMNS = [
    "xref",
    "given_name",
    "surname",
    "suffix",
    "sex",
    "birth_date",
    "birth_year",
    "birth_place",
    "death_date",
    "death_year",
    "death_place",
    "burial_date",
    "burial_place",
    "occupations",
    "source_count",
    "famc_xref",
    "fams_xrefs",
]

_FAM_CSV_COLUMNS = [
    "xref",
    "husband_xref",
    "husband_name",
    "wife_xref",
    "wife_name",
    "marriage_date",
    "marriage_year",
    "marriage_place",
    "child_count",
    "children_xrefs",
]


def _redact_individual_csv(indi: ExportIndividual) -> list[str]:
    """Build a redacted CSV row for a living individual."""
    return [
        indi.xref,
        "Living",
        "",  # surname
        "",  # suffix
        indi.sex,
        "",  # birth_date
        "",  # birth_year
        "",  # birth_place
        "",  # death_date
        "",  # death_year
        "",  # death_place
        "",  # burial_date
        "",  # burial_place
        "",  # occupations
        str(indi.source_count),
        indi.famc_xref,
        ";".join(indi.fams_xrefs),
    ]


def _individual_csv_row(indi: ExportIndividual) -> list[str]:
    return [
        indi.xref,
        indi.given_name,
        indi.surname,
        indi.suffix,
        indi.sex,
        indi.birth_date,
        str(indi.birth_year) if indi.birth_year is not None else "",
        indi.birth_place,
        indi.death_date,
        str(indi.death_year) if indi.death_year is not None else "",
        indi.death_place,
        indi.burial_date,
        indi.burial_place,
        "; ".join(indi.occupations),
        str(indi.source_count),
        indi.famc_xref,
        ";".join(indi.fams_xrefs),
    ]


def _family_csv_row(
    fam: ExportFamily, living_xrefs: set[str] | None = None
) -> list[str]:
    husband_name = fam.husband_name
    wife_name = fam.wife_name
    if living_xrefs:
        if fam.husband_xref in living_xrefs:
            husband_name = "Living"
        if fam.wife_xref in living_xrefs:
            wife_name = "Living"
    return [
        fam.xref,
        fam.husband_xref,
        husband_name,
        fam.wife_xref,
        wife_name,
        fam.marriage_date,
        str(fam.marriage_year) if fam.marriage_year is not None else "",
        fam.marriage_place,
        str(fam.child_count),
        ";".join(fam.children_xrefs),
    ]


def format_csv(
    result: ExportResult,
    table: str = "individuals",
    include_bom: bool = True,
    redact_living: bool = False,
    max_age: int = 110,
) -> str:
    buf = io.StringIO()
    if include_bom:
        buf.write("\ufeff")

    writer = csv.writer(buf)

    living_xrefs: set[str] | None = None
    if redact_living:
        living_xrefs = {
            indi.xref
            for indi in result.individuals
            if estimate_living(
                indi.birth_year, indi.death_year, indi.burial_date, max_age=max_age
            )
        }

    if table == "families":
        writer.writerow(_FAM_CSV_COLUMNS)
        for fam in result.families:
            writer.writerow(_family_csv_row(fam, living_xrefs))
    else:
        writer.writerow(_INDI_CSV_COLUMNS)
        for indi in result.individuals:
            if living_xrefs and indi.xref in living_xrefs:
                writer.writerow(_redact_individual_csv(indi))
            else:
                writer.writerow(_individual_csv_row(indi))

    return buf.getvalue()


def _individual_to_dict(
    indi: ExportIndividual, redacted: bool = False
) -> dict[str, Any]:
    if redacted:
        return {
            "xref": indi.xref,
            "given_name": "Living",
            "surname": "",
            "suffix": "",
            "sex": indi.sex,
            "birth_date": "",
            "birth_year": None,
            "birth_place": "",
            "death_date": "",
            "death_year": None,
            "death_place": "",
            "burial_date": "",
            "burial_place": "",
            "occupations": [],
            "source_count": indi.source_count,
            "famc_xref": indi.famc_xref,
            "fams_xrefs": list(indi.fams_xrefs),
            "alt_names": [],
            "notes": [],
        }
    return {
        "xref": indi.xref,
        "given_name": indi.given_name,
        "surname": indi.surname,
        "suffix": indi.suffix,
        "sex": indi.sex,
        "birth_date": indi.birth_date,
        "birth_year": indi.birth_year,
        "birth_place": indi.birth_place,
        "death_date": indi.death_date,
        "death_year": indi.death_year,
        "death_place": indi.death_place,
        "burial_date": indi.burial_date,
        "burial_place": indi.burial_place,
        "occupations": list(indi.occupations),
        "source_count": indi.source_count,
        "famc_xref": indi.famc_xref,
        "fams_xrefs": list(indi.fams_xrefs),
        "alt_names": [{"given": g, "surname": s} for g, s in indi.alt_names],
        "notes": list(indi.notes),
    }


def _family_to_dict(
    fam: ExportFamily, living_xrefs: set[str] | None = None
) -> dict[str, Any]:
    husband_name = fam.husband_name
    wife_name = fam.wife_name
    if living_xrefs:
        if fam.husband_xref in living_xrefs:
            husband_name = "Living"
        if fam.wife_xref in living_xrefs:
            wife_name = "Living"
    return {
        "xref": fam.xref,
        "husband_xref": fam.husband_xref,
        "husband_name": husband_name,
        "wife_xref": fam.wife_xref,
        "wife_name": wife_name,
        "marriage_date": fam.marriage_date,
        "marriage_year": fam.marriage_year,
        "marriage_place": fam.marriage_place,
        "child_count": fam.child_count,
        "children_xrefs": list(fam.children_xrefs),
    }


def format_json(
    result: ExportResult,
    redact_living: bool = False,
    max_age: int = 110,
) -> str:
    living_xrefs: set[str] | None = None
    if redact_living:
        living_xrefs = {
            indi.xref
            for indi in result.individuals
            if estimate_living(
                indi.birth_year, indi.death_year, indi.burial_date, max_age=max_age
            )
        }

    data: dict[str, Any] = {
        "meta": {
            "file": result.file_path,
            "encoding": result.encoding,
            "gedcom_tools_version": __version__,
            "individual_count": result.individual_count,
            "family_count": result.family_count,
            "redacted_living": redact_living,
        },
        "individuals": [
            _individual_to_dict(
                indi, redacted=living_xrefs is not None and indi.xref in living_xrefs
            )
            for indi in result.individuals
        ],
        "families": [_family_to_dict(fam, living_xrefs) for fam in result.families],
    }

    return json.dumps(data, indent=2, ensure_ascii=False)



from __future__ import annotations

from gedcom_tools.commands.compare.models import (
    CompareIndividual,
    CompareResult,
    FieldDiff,
    MatchPair,
    MatchScore,
)
from gedcom_tools.utils import EncodingInfo


class TestCompareIndividual:

    def test_defaults(self) -> None:
        ind = CompareIndividual(xref="@I1@", source_file="A")
        assert ind.xref == "@I1@"
        assert ind.source_file == "A"
        assert ind.given_name == ""
        assert ind.surname == ""
        assert ind.full_name == ""
        assert ind.sex == ""
        assert ind.birth_year is None
        assert ind.birth_place == ""
        assert ind.death_year is None
        assert ind.death_place == ""
        assert ind.famc_xref is None
        assert ind.fams_xrefs == []
        assert ind.alt_surnames == []
        assert ind.alt_given_names == []
        assert ind.given_name_normalized == ""
        assert ind.surname_normalized == ""
        assert ind.birth_place_normalized == ""
        assert ind.death_place_normalized == ""
        assert ind.alt_surnames_normalized == []
        assert ind.alt_given_names_normalized == []
        assert ind.surname_phonetic == ""
        assert ind.given_phonetic == ""
        assert ind.birth_decade == ""
        assert ind.death_decade == ""

    def test_full_construction(self) -> None:
        ind = CompareIndividual(
            xref="@I55@",
            source_file="B",
            given_name="Eleni",
            surname="Papadopoulos",
            full_name="Eleni Papadopoulos",
            sex="F",
            birth_year=1852,
            birth_place="Athens",
            death_year=1920,
            death_place="Patras",
            famc_xref="@F3@",
            fams_xrefs=["@F10@", "@F11@"],
            alt_surnames=["Papadopoulou"],
            alt_given_names=["Helen"],
            given_name_normalized="eleni",
            surname_normalized="papadopoulos",
            birth_place_normalized="athens",
            death_place_normalized="patras",
            alt_surnames_normalized=["papadopoulou"],
            alt_given_names_normalized=["helen"],
            surname_phonetic="P131",
            given_phonetic="E450",
            birth_decade="1850s",
            death_decade="1920s",
        )
        assert ind.given_name == "Eleni"
        assert ind.surname == "Papadopoulos"
        assert ind.birth_year == 1852
        assert ind.fams_xrefs == ["@F10@", "@F11@"]
        assert ind.alt_surnames == ["Papadopoulou"]
        assert ind.surname_phonetic == "P131"
        assert ind.birth_decade == "1850s"

    def test_list_fields_independent(self) -> None:
        a = CompareIndividual(xref="@I1@", source_file="A")
        b = CompareIndividual(xref="@I2@", source_file="B")
        a.fams_xrefs.append("@F1@")
        a.alt_surnames.append("Doe")
        a.alt_given_names.append("Johnny")
        a.alt_surnames_normalized.append("doe")
        a.alt_given_names_normalized.append("johnny")
        assert b.fams_xrefs == []
        assert b.alt_surnames == []
        assert b.alt_given_names == []
        assert b.alt_surnames_normalized == []
        assert b.alt_given_names_normalized == []


class TestMatchScore:

    def test_defaults(self) -> None:
        score = MatchScore(
            total=0.85, field_scores={"surname": 1.0}, classification="probable"
        )
        assert score.total == 0.85
        assert score.field_scores == {"surname": 1.0}
        assert score.classification == "probable"
        assert score.insufficient_data is False
        assert score.name_only is False
        assert score.comparable_field_count == 0

    def test_classification_stored(self) -> None:
        for label in ("certain", "probable", "non_match"):
            score = MatchScore(total=0.5, field_scores={}, classification=label)
            assert score.classification == label


class TestFieldDiff:

    def test_construction(self) -> None:
        diff = FieldDiff(field="Birth Year", value_a="1852", value_b="1853")
        assert diff.field == "Birth Year"
        assert diff.value_a == "1852"
        assert diff.value_b == "1853"


class TestMatchPair:

    def test_construction(self) -> None:
        ind_a = CompareIndividual(xref="@I1@", source_file="A", full_name="John Smith")
        ind_b = CompareIndividual(xref="@I5@", source_file="B", full_name="John Smyth")
        score = MatchScore(
            total=0.78,
            field_scores={"surname": 0.8, "given_name": 1.0},
            classification="probable",
            comparable_field_count=2,
        )
        diffs = [FieldDiff(field="Surname", value_a="Smith", value_b="Smyth")]
        pair = MatchPair(
            individual_a=ind_a,
            individual_b=ind_b,
            score=score,
            field_diffs=diffs,
        )
        assert pair.individual_a.xref == "@I1@"
        assert pair.individual_b.full_name == "John Smyth"
        assert pair.score.total == 0.78
        assert len(pair.field_diffs) == 1
        assert pair.field_diffs[0].field == "Surname"


class TestCompareResult:

    def test_construction(self) -> None:
        enc_a = EncodingInfo(encoding="UTF-8", has_bom=False)
        enc_b = EncodingInfo(encoding="ANSEL", declared_charset="ANSEL")
        ind_a = CompareIndividual(xref="@I1@", source_file="A")
        ind_b = CompareIndividual(xref="@I2@", source_file="B")
        score = MatchScore(
            total=0.95,
            field_scores={"surname": 1.0, "birth_year": 1.0},
            classification="certain",
            comparable_field_count=2,
        )
        pair = MatchPair(
            individual_a=ind_a,
            individual_b=ind_b,
            score=score,
            field_diffs=[],
        )
        result = CompareResult(
            file_a="tree_a.ged",
            file_b="tree_b.ged",
            encoding_a=enc_a,
            encoding_b=enc_b,
            total_a=150,
            total_b=200,
            certain_matches=[pair],
            probable_matches=[],
            unique_to_a=[CompareIndividual(xref="@I99@", source_file="A")],
            unique_to_b=[],
        )
        assert result.file_a == "tree_a.ged"
        assert result.encoding_b.encoding == "ANSEL"
        assert result.total_a == 150
        assert result.total_b == 200
        assert len(result.certain_matches) == 1
        assert result.probable_matches == []
        assert len(result.unique_to_a) == 1
        assert result.unique_to_b == []



"""Transform operations for the filter command.

Applies strip operations to GEDCOM records: removing top-level records
by tag, removing sub-lines by tag (with skip-depth for children),
stripping custom (underscore-prefixed) tags at any level, and
extracting subtrees rooted at a specific individual.
"""

from __future__ import annotations

from pathlib import Path

from gedcom_tools.commands.filter.models import (
    UNLIMITED_DEPTH,
    FilterSpec,
    GedcomLine,
    GedcomRecord,
)
from gedcom_tools.commands.filter.parser import (
    build_xref_tag_map,
    collect_xrefs,
    is_pointer_value,
)
from gedcom_tools.graph import (
    build_parent_child_graph,
    find_ancestors,
    find_descendants,
)


def apply_strip_transforms(
    records: list[GedcomRecord], spec: FilterSpec
) -> tuple[list[GedcomRecord], set[str]]:
    """Apply all strip operations defined in the filter spec.

    Returns filtered records and the set of xrefs removed at the
    record level (needed for downstream dangling-pointer cleanup).
    """
    all_removed_xrefs: set[str] = set()

    # Collect tags for record-level and line-level stripping
    record_tags: set[str] = set()
    line_tags: set[str] = set()

    if spec.strip_notes:
        record_tags.add("NOTE")
        line_tags.add("NOTE")
    if spec.strip_sources:
        record_tags.add("SOUR")
        line_tags.add("SOUR")
    if spec.strip_multimedia:
        record_tags.add("OBJE")
        line_tags.add("OBJE")

    # 1. Record-level strips (remove entire NOTE/SOUR/OBJE records)
    if record_tags:
        records, removed = _strip_records_by_tag(records, record_tags)
        all_removed_xrefs.update(removed)

    # 2. Custom line strips
    if spec.strip_custom_tags:
        records = _strip_custom_lines(records)

    # 3. Line-level strips (remove inline NOTE/SOUR/OBJE sub-tags)
    if line_tags:
        records = _strip_lines_by_tag(records, line_tags)

    # 4. Strip-tag (user-specified tags, both record and line level)
    if spec.strip_tags:
        user_tags = {t.upper() for t in spec.strip_tags}
        records, removed = _strip_records_by_tag(records, user_tags)
        all_removed_xrefs.update(removed)
        records = _strip_lines_by_tag(records, user_tags)

    return records, all_removed_xrefs


def _strip_records_by_tag(
    records: list[GedcomRecord], tags: set[str]
) -> tuple[list[GedcomRecord], set[str]]:
    """Remove level-0 records whose tag is in the given set.

    Returns the filtered record list and the set of xrefs that were
    removed (for dangling-pointer cleanup downstream).
    """
    kept: list[GedcomRecord] = []
    removed_xrefs: set[str] = set()

    for rec in records:
        if rec.tag in tags:
            if rec.xref is not None:
                removed_xrefs.add(rec.xref)
            continue
        kept.append(rec)

    return kept, removed_xrefs


def _strip_lines_by_tag(
    records: list[GedcomRecord], tags: set[str]
) -> list[GedcomRecord]:
    """Remove sub-lines (level 1+) whose tag matches, including children.

    Uses skip-depth logic: when a line at level N is removed, all
    subsequent lines at level > N are also skipped until reaching a
    line at level <= N. Does not mutate the original records.
    """
    result: list[GedcomRecord] = []

    for rec in records:
        filtered_children: list[GedcomLine] = []
        skip_below: int | None = None

        for child in rec.children:
            if skip_below is not None:
                if child.level > skip_below:
                    continue
                skip_below = None

            if child.tag in tags:
                skip_below = child.level
                continue

            filtered_children.append(child)

        result.append(GedcomRecord(header=rec.header, children=filtered_children))

    return result


def _strip_custom_lines(records: list[GedcomRecord]) -> list[GedcomRecord]:
    """Remove lines whose tag starts with ``_`` at any level, including children.

    Uses the same skip-depth logic as ``_strip_lines_by_tag``: when a
    custom tag at level N is removed, all subsequent lines at level > N
    are also skipped. Does not mutate the original records.
    """
    result: list[GedcomRecord] = []

    for rec in records:
        filtered_children: list[GedcomLine] = []
        skip_below: int | None = None

        for child in rec.children:
            if skip_below is not None:
                if child.level > skip_below:
                    continue
                skip_below = None

            if child.tag.startswith("_"):
                skip_below = child.level
                continue

            filtered_children.append(child)

        result.append(GedcomRecord(header=rec.header, children=filtered_children))

    return result


# ---------------------------------------------------------------------------
# Subtree extraction
# ---------------------------------------------------------------------------


# Tags that are always preserved regardless of keep set
_STRUCTURAL_TAGS = frozenset({"HEAD", "TRLR", "SUBM"})

# Record types eligible for transitive dependency collection
_DEPENDENT_TAGS = frozenset({"SOUR", "NOTE", "OBJE", "REPO"})


def _family_has_kept_member(record: GedcomRecord, kept_xrefs: set[str]) -> bool:
    """Check if a FAM record has any HUSB, WIFE, or CHIL in the keep set."""
    for child in record.children:
        if child.tag in ("HUSB", "WIFE", "CHIL"):
            if is_pointer_value(child.value) and child.value is not None:
                stripped = child.value.strip()
                if stripped in kept_xrefs:
                    return True
    return False


def _collect_dependent_xrefs(
    records: list[GedcomRecord],
    kept_xrefs: set[str],
    xref_to_tag: dict[str, str],
) -> set[str]:
    """Transitively collect SOUR/NOTE/OBJE/REPO xrefs referenced by kept records.

    Scans children of records whose xref is in kept_xrefs. For each
    pointer value pointing to a SOUR/NOTE/OBJE/REPO record (checked via
    xref_to_tag), adds it to the result set. Repeats until no new xrefs
    are discovered, ensuring chains like SOUR -> REPO are preserved.
    """
    # Build lookup for fast record access by xref
    rec_by_xref: dict[str, GedcomRecord] = {}
    for rec in records:
        if rec.xref is not None:
            rec_by_xref[rec.xref] = rec

    collected: set[str] = set()
    # Seed: scan all currently-kept records
    to_scan: set[str] = set(kept_xrefs)
    scanned: set[str] = set()

    while to_scan:
        current_xref = to_scan.pop()
        scanned.add(current_xref)

        current_rec = rec_by_xref.get(current_xref)
        if current_rec is None:
            continue

        for child in current_rec.children:
            if not is_pointer_value(child.value) or child.value is None:
                continue
            target = child.value.strip()
            target_tag = xref_to_tag.get(target)
            if target_tag not in _DEPENDENT_TAGS:
                continue
            if target in collected:
                continue
            collected.add(target)
            if target not in scanned:
                to_scan.add(target)

    return collected


def extract_subtree(
    records: list[GedcomRecord],
    file_path: Path,
    root_xref: str,
    ancestor_depth: int | None,
    descendant_depth: int,
    include_spouses: bool,
) -> tuple[list[GedcomRecord], set[str]]:
    """Extract a subtree centered on an individual.

    Keeps the root individual plus ancestors (up to ancestor_depth),
    descendants (up to descendant_depth), optionally their spouses,
    related FAM records, and transitively referenced SOUR/NOTE/OBJE/REPO
    records. HEAD, TRLR, and SUBM records are always preserved.

    Returns the filtered record list and the set of removed xrefs.
    """
    graph = build_parent_child_graph(file_path)

    all_xrefs = collect_xrefs(records)
    if root_xref not in all_xrefs:
        raise ValueError(f"Individual {root_xref} not found in file")

    # Build the individual keep set
    keep: set[str] = {root_xref}

    # Ancestors
    resolved_ancestor_depth = (
        UNLIMITED_DEPTH if ancestor_depth is None else ancestor_depth
    )
    if resolved_ancestor_depth > 0:
        keep |= find_ancestors(graph, root_xref, resolved_ancestor_depth)

    # Descendants
    if descendant_depth > 0:
        keep |= find_descendants(graph, root_xref, descendant_depth)

    # Spouses of all kept individuals
    if include_spouses:
        spouse_xrefs: set[str] = set()
        for xref in keep:
            spouse_xrefs |= graph.couples.get(xref, set())
        keep |= spouse_xrefs

    # FAM records: keep any where HUSB, WIFE, or CHIL is in the keep set
    xref_to_tag = build_xref_tag_map(records)
    fam_keep: set[str] = set()
    for rec in records:
        if rec.tag == "FAM" and rec.xref is not None:
            if _family_has_kept_member(rec, keep):
                fam_keep.add(rec.xref)

    full_keep = keep | fam_keep

    # Transitive dependent records (SOUR/NOTE/OBJE/REPO chains)
    dependents = _collect_dependent_xrefs(records, full_keep, xref_to_tag)
    full_keep |= dependents

    # Filter records
    kept_records: list[GedcomRecord] = []
    removed_xrefs: set[str] = set()

    for rec in records:
        if rec.tag in _STRUCTURAL_TAGS:
            kept_records.append(rec)
        elif rec.xref is None:
            # Non-xref structural records (rare, but preserve them)
            kept_records.append(rec)
        elif rec.xref in full_keep:
            kept_records.append(rec)
        else:
            removed_xrefs.add(rec.xref)

    return kept_records, removed_xrefs



"""Stats command for GEDCOM files."""

from __future__ import annotations

import argparse
import sys
from pathlib import Path
from typing import TYPE_CHECKING

# Re-export public API
from gedcom_tools.commands.stats.collector import StatsCollector
from gedcom_tools.commands.stats.formatters import StatsResult
from gedcom_tools.commands.stats.models import (
    AggregateStats,
    CoverageStats,
    DatePrecisionStats,
    FamilyData,
    FamilyEntry,
    GenderedAggregateStats,
    GenerationEntry,
    IndividualData,
    LifespanStats,
    MarriageStats,
    RankedItem,
    TimelineEntry,
)
from gedcom_tools.constants import EXIT_ERROR, EXIT_SUCCESS
from gedcom_tools.progress import Colors
from gedcom_tools.utils import validate_input_file

if TYPE_CHECKING:
    from argparse import Namespace, _SubParsersAction

__all__ = [
    # Main API
    "register_subcommand",
    "run",
    "StatsCollector",
    "StatsResult",
    # Models
    "AggregateStats",
    "CoverageStats",
    "DatePrecisionStats",
    "FamilyData",
    "FamilyEntry",
    "GenderedAggregateStats",
    "GenerationEntry",
    "IndividualData",
    "LifespanStats",
    "MarriageStats",
    "RankedItem",
    "TimelineEntry",
]


def register_subcommand(subparsers: _SubParsersAction[argparse.ArgumentParser]) -> None:
    """Register the stats subcommand."""
    parser = subparsers.add_parser(
        "stats",
        help="Display statistics about a GEDCOM file",
        description="Analyze a GEDCOM file and display genealogical statistics.",
    )
    parser.add_argument(
        "file",
        type=Path,
        help="Path to the GEDCOM file to analyze",
    )
    parser.add_argument(
        "--top",
        type=int,
        default=10,
        metavar="N",
        help="Number of items in top-N lists (default: 10)",
    )


def run(args: Namespace) -> int:
    """Execute the stats command."""
    file_path: Path = args.file

    if err := validate_input_file(file_path):
        return err

    output_format = getattr(args, "format", "text")
    quiet = getattr(args, "quiet", False)
    verbose = getattr(args, "verbose", False)
    no_color = getattr(args, "no_color", False)
    top_n = getattr(args, "top", 10)

    try:
        collector = StatsCollector(
            file_path=file_path,
            quiet=quiet,
            verbose=verbose,
            no_color=no_color,
            top_n=top_n,
        )
        result = collector.collect()

        if output_format == "json":
            print(result.format_json())
        else:
            colors = Colors(sys.stdout, force_disable=no_color)
            output = result.format_text(colors, quiet=quiet)
            if output:
                print(output)

        return EXIT_SUCCESS

    except Exception as e:
        if verbose:
            raise
        print(f"Error: {e}", file=sys.stderr)
        return EXIT_ERROR



from __future__ import annotations

import argparse
import json
from pathlib import Path

from gedcom_tools.commands.compare import register_subcommand, run
from gedcom_tools.constants import EXIT_SUCCESS, EXIT_USAGE_ERROR


def _write_gedcom(path: Path, individuals: list[str]) -> Path:
    lines = [
        "0 HEAD",
        "1 SOUR TEST",
        "1 GEDC",
        "2 VERS 5.5.1",
        "1 CHAR UTF-8",
    ]
    for indi in individuals:
        lines.append(indi)
    lines.append("0 TRLR")
    path.write_text("\n".join(lines) + "\n", encoding="utf-8")
    return path


def _indi_block(
    xref: str,
    given: str = "",
    surname: str = "",
    sex: str = "",
    birth_year: int | None = None,
    death_year: int | None = None,
    birth_place: str = "",
    death_place: str = "",
) -> str:
    lines = [f"0 {xref} INDI"]
    if given or surname:
        lines.append(f"1 NAME {given} /{surname}/")
    if sex:
        lines.append(f"1 SEX {sex}")
    if birth_year is not None or birth_place:
        lines.append("1 BIRT")
        if birth_year is not None:
            lines.append(f"2 DATE {birth_year}")
        if birth_place:
            lines.append(f"2 PLAC {birth_place}")
    if death_year is not None or death_place:
        lines.append("1 DEAT")
        if death_year is not None:
            lines.append(f"2 DATE {death_year}")
        if death_place:
            lines.append(f"2 PLAC {death_place}")
    return "\n".join(lines)


def _make_args(
    file_a: Path,
    file_b: Path,
    fmt: str = "text",
    quiet: bool = False,
    verbose: bool = False,
    no_color: bool = True,
    certain_threshold: float = 0.85,
    probable_threshold: float = 0.65,
    show_matches: str = "all",
    list_unique: bool = False,
    limit: int | None = None,
    reject_sex_mismatch: bool = False,
    phonetic: str = "soundex",
) -> argparse.Namespace:
    return argparse.Namespace(
        file_a=file_a,
        file_b=file_b,
        format=fmt,
        quiet=quiet,
        verbose=verbose,
        no_color=no_color,
        certain_threshold=certain_threshold,
        probable_threshold=probable_threshold,
        show_matches=show_matches,
        list_unique=list_unique,
        limit=limit,
        reject_sex_mismatch=reject_sex_mismatch,
        phonetic=phonetic,
    )


# Shared individual blocks for File A
_FILE_A_INDIVIDUALS = [
    _indi_block(
        "@I1@",
        "John",
        "Smith",
        "M",
        birth_year=1850,
        death_year=1920,
        birth_place="London, England",
    ),
    _indi_block(
        "@I2@",
        "Mary",
        "Johnson",
        "F",
        birth_year=1872,
        death_year=1945,
    ),
    _indi_block(
        "@I3@",
        "Robert",
        "Williams",
        "M",
        birth_year=1900,
    ),
]

# Shared individual blocks for File B
_FILE_B_INDIVIDUALS = [
    _indi_block(
        "@I10@",
        "John",
        "Smith",
        "M",
        birth_year=1850,
        death_year=1920,
        birth_place="London, Middlesex, England",
    ),
    _indi_block(
        "@I11@",
        "Mary",
        "Johnson",
        "F",
        birth_year=1872,
        death_year=1945,
    ),
    _indi_block(
        "@I12@",
        "Alice",
        "Brown",
        "F",
        birth_year=1830,
    ),
]


def _create_pair(tmp_path: Path) -> tuple[Path, Path]:
    file_a = _write_gedcom(tmp_path / "a.ged", _FILE_A_INDIVIDUALS)
    file_b = _write_gedcom(tmp_path / "b.ged", _FILE_B_INDIVIDUALS)
    return file_a, file_b


class TestRegisterSubcommand:
    def test_parser_created(self) -> None:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers()
        register_subcommand(subparsers)

        args = parser.parse_args(
            [
                "compare",
                "a.ged",
                "b.ged",
                "--certain-threshold",
                "0.90",
                "--probable-threshold",
                "0.70",
                "--show-matches",
                "certain",
                "--list-unique",
                "--limit",
                "10",
                "--reject-sex-mismatch",
            ]
        )
        assert args.file_a == Path("a.ged")
        assert args.file_b == Path("b.ged")
        assert args.certain_threshold == 0.90
        assert args.probable_threshold == 0.70
        assert args.show_matches == "certain"
        assert args.list_unique is True
        assert args.limit == 10
        assert args.reject_sex_mismatch is True


class TestValidation:
    def test_missing_file_a(self, tmp_path: Path) -> None:
        missing = tmp_path / "nope.ged"
        real = _write_gedcom(tmp_path / "b.ged", [])
        args = _make_args(missing, real)
        assert run(args) == EXIT_USAGE_ERROR

    def test_missing_file_b(self, tmp_path: Path) -> None:
        real = _write_gedcom(tmp_path / "a.ged", [])
        missing = tmp_path / "nope.ged"
        args = _make_args(real, missing)
        assert run(args) == EXIT_USAGE_ERROR

    def test_same_file_detection(self, tmp_path: Path, capsys) -> None:
        f = _write_gedcom(tmp_path / "same.ged", [])
        args = _make_args(f, f)
        assert run(args) == EXIT_USAGE_ERROR
        assert "same file" in capsys.readouterr().err.lower()

    def test_invalid_certain_threshold(self, tmp_path: Path) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, certain_threshold=1.5)
        assert run(args) == EXIT_USAGE_ERROR

    def test_invalid_probable_threshold(self, tmp_path: Path) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, probable_threshold=-0.1)
        assert run(args) == EXIT_USAGE_ERROR

    def test_certain_not_greater_than_probable(self, tmp_path: Path) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, certain_threshold=0.50, probable_threshold=0.65)
        assert run(args) == EXIT_USAGE_ERROR


class TestBasicPipeline:
    def test_successful_comparison(self, tmp_path: Path) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b)
        assert run(args) == EXIT_SUCCESS

    def test_text_output_has_summary(self, tmp_path: Path, capsys) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, fmt="text")
        run(args)
        assert "=== Comparison Summary ===" in capsys.readouterr().out

    def test_json_output_valid(self, tmp_path: Path, capsys) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, fmt="json")
        run(args)
        data = json.loads(capsys.readouterr().out)
        assert "certain_matches" in data
        assert "probable_matches" in data
        assert "unique_to_a" in data
        assert "unique_to_b" in data
        assert "total_a" in data
        assert "total_b" in data

    def test_quiet_output_single_line(self, tmp_path: Path, capsys) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, quiet=True)
        run(args)
        out = capsys.readouterr().out.strip()
        assert "\n" not in out
        assert "certain" in out

    def test_matching_individuals_found(self, tmp_path: Path, capsys) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, fmt="json")
        run(args)
        data = json.loads(capsys.readouterr().out)

        # John Smith (I1 <-> I10) and Mary Johnson (I2 <-> I11) should match
        all_matches = data["certain_matches"] + data["probable_matches"]
        matched_pairs = {
            (m["individual_a"]["xref"], m["individual_b"]["xref"]) for m in all_matches
        }
        assert ("@I1@", "@I10@") in matched_pairs
        assert ("@I2@", "@I11@") in matched_pairs


class TestFlags:
    def test_show_certain_only(self, tmp_path: Path, capsys) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, fmt="json", show_matches="certain")
        run(args)
        data = json.loads(capsys.readouterr().out)
        assert data["probable_matches"] == []

    def test_show_probable_only(self, tmp_path: Path, capsys) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, fmt="json", show_matches="probable")
        run(args)
        data = json.loads(capsys.readouterr().out)
        assert data["certain_matches"] == []

    def test_list_unique(self, tmp_path: Path, capsys) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, fmt="json", list_unique=True)
        run(args)
        data = json.loads(capsys.readouterr().out)
        unique_a_xrefs = {u["xref"] for u in data["unique_to_a"]}
        unique_b_xrefs = {u["xref"] for u in data["unique_to_b"]}
        # Robert Williams only in A, Alice Brown only in B
        assert "@I3@" in unique_a_xrefs
        assert "@I12@" in unique_b_xrefs

    def test_limit_truncates(self, tmp_path: Path, capsys) -> None:
        # Build files with many overlapping individuals
        indis_a = [
            _indi_block(
                f"@I{i}@",
                f"Person{i}",
                "Commonname",
                "M",
                birth_year=1800 + i,
                death_year=1870 + i,
            )
            for i in range(1, 11)
        ]
        indis_b = [
            _indi_block(
                f"@J{i}@",
                f"Person{i}",
                "Commonname",
                "M",
                birth_year=1800 + i,
                death_year=1870 + i,
            )
            for i in range(1, 11)
        ]
        fa = _write_gedcom(tmp_path / "a.ged", indis_a)
        fb = _write_gedcom(tmp_path / "b.ged", indis_b)
        args = _make_args(fa, fb, fmt="json", limit=1)
        run(args)
        data = json.loads(capsys.readouterr().out)
        # limit=1 caps each section to 1 item
        assert data["certain_matches"] == [] or len(data["certain_matches"]) <= 1
        assert data["probable_matches"] == [] or len(data["probable_matches"]) <= 1


class TestLimitDefaulting:
    def test_text_default_limit(self, tmp_path: Path, capsys) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, fmt="text", limit=None)
        rc = run(args)
        assert rc == EXIT_SUCCESS
        # No crash — text limit defaults to 50 internally

    def test_json_default_limit(self, tmp_path: Path, capsys) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, fmt="json", limit=None)
        run(args)
        data = json.loads(capsys.readouterr().out)
        # JSON default is unlimited — all items present
        all_matches = data["certain_matches"] + data["probable_matches"]
        matched_xrefs = {m["individual_a"]["xref"] for m in all_matches}
        # Both matching individuals should be present (not truncated)
        assert "@I1@" in matched_xrefs and "@I2@" in matched_xrefs


class TestExitCodes:
    def test_success_exit_code(self, tmp_path: Path) -> None:
        a, b = _create_pair(tmp_path)
        assert run(_make_args(a, b)) == EXIT_SUCCESS

    def test_file_not_found_exit_code(self, tmp_path: Path) -> None:
        missing = tmp_path / "gone.ged"
        real = _write_gedcom(tmp_path / "b.ged", [])
        assert run(_make_args(missing, real)) == EXIT_USAGE_ERROR

    def test_threshold_error_exit_code(self, tmp_path: Path) -> None:
        a, b = _create_pair(tmp_path)
        args = _make_args(a, b, certain_threshold=0.50, probable_threshold=0.65)
        assert run(args) == EXIT_USAGE_ERROR


class TestEdgeCases:
    def test_empty_files(self, tmp_path: Path, capsys) -> None:
        fa = _write_gedcom(tmp_path / "a.ged", [])
        fb = _write_gedcom(tmp_path / "b.ged", [])
        args = _make_args(fa, fb, fmt="json")
        rc = run(args)
        assert rc == EXIT_SUCCESS
        data = json.loads(capsys.readouterr().out)
        assert data["certain_matches"] == []
        assert data["probable_matches"] == []
        assert data["total_a"] == 0
        assert data["total_b"] == 0

    def test_no_overlap(self, tmp_path: Path, capsys) -> None:
        fa = _write_gedcom(
            tmp_path / "a.ged",
            [
                _indi_block(
                    "@I1@", "Unique", "Alpha", "M", birth_year=1800, death_year=1870
                )
            ],
        )
        fb = _write_gedcom(
            tmp_path / "b.ged",
            [
                _indi_block(
                    "@I1@", "Different", "Omega", "F", birth_year=1950, death_year=2020
                )
            ],
        )
        args = _make_args(fa, fb, fmt="json", list_unique=True)
        rc = run(args)
        assert rc == EXIT_SUCCESS
        data = json.loads(capsys.readouterr().out)
        assert data["certain_matches"] == []
        assert data["probable_matches"] == []
        assert len(data["unique_to_a"]) == 1
        assert len(data["unique_to_b"]) == 1

    def test_identical_individuals(self, tmp_path: Path, capsys) -> None:
        block = _indi_block(
            "@I1@",
            "George",
            "Washington",
            "M",
            birth_year=1732,
            death_year=1799,
            birth_place="Westmoreland, Virginia",
            death_place="Mount Vernon, Virginia",
        )
        fa = _write_gedcom(tmp_path / "a.ged", [block])
        # Same data, different xref in file B
        block_b = _indi_block(
            "@I99@",
            "George",
            "Washington",
            "M",
            birth_year=1732,
            death_year=1799,
            birth_place="Westmoreland, Virginia",
            death_place="Mount Vernon, Virginia",
        )
        fb = _write_gedcom(tmp_path / "b.ged", [block_b])
        args = _make_args(fa, fb, fmt="json")
        rc = run(args)
        assert rc == EXIT_SUCCESS
        data = json.loads(capsys.readouterr().out)
        assert len(data["certain_matches"]) == 1
        pair = data["certain_matches"][0]
        assert pair["individual_a"]["xref"] == "@I1@"
        assert pair["individual_b"]["xref"] == "@I99@"
        assert pair["score"] >= 0.85


class TestMetaphoneIntegration:
    def test_metaphone_option_accepted(self, tmp_path: Path, capsys) -> None:
        fa = _write_gedcom(
            tmp_path / "a.ged",
            [
                _indi_block(
                    "@I1@",
                    "John",
                    "Smith",
                    "M",
                    birth_year=1850,
                    death_year=1920,
                    birth_place="London, England",
                ),
            ],
        )
        fb = _write_gedcom(
            tmp_path / "b.ged",
            [
                _indi_block(
                    "@I2@",
                    "John",
                    "Smith",
                    "M",
                    birth_year=1850,
                    death_year=1920,
                    birth_place="London, England",
                ),
            ],
        )
        args = _make_args(fa, fb, fmt="json", phonetic="metaphone")
        rc = run(args)
        assert rc == EXIT_SUCCESS
        data = json.loads(capsys.readouterr().out)
        assert len(data["certain_matches"]) == 1

    def test_metaphone_cross_code_blocking(self, tmp_path: Path, capsys) -> None:
        # Use surnames with different soundex first letters but shared metaphone.
        # Only blocking key is surname+birth — no given names, different death years,
        # no exact year combo.
        fa = _write_gedcom(
            tmp_path / "a.ged",
            [
                _indi_block(
                    "@I1@",
                    "John",
                    "Catherine",
                    "M",
                    birth_year=1850,
                    birth_place="London",
                ),
            ],
        )
        fb = _write_gedcom(
            tmp_path / "b.ged",
            [
                _indi_block(
                    "@I2@",
                    "James",
                    "Katherine",
                    "M",
                    birth_year=1853,
                    birth_place="London",
                ),
            ],
        )
        # Soundex: surname Catherine(C365) vs Katherine(K365) — different.
        # Given names differ too (J500 vs J520). No death year combo.
        # Only possible blocking pass: surname+birth decade.
        # Soundex surname codes differ → no candidates → no match.
        args_sx = _make_args(fa, fb, fmt="json", phonetic="soundex")
        run(args_sx)
        data_sx = json.loads(capsys.readouterr().out)
        assert data_sx["certain_matches"] + data_sx["probable_matches"] == []

        # Metaphone: Catherine and Katherine produce same codes.
        # Multi-key blocking catches them via shared code + birth decade.
        args_mp = _make_args(fa, fb, fmt="json", phonetic="metaphone")
        run(args_mp)
        data_mp = json.loads(capsys.readouterr().out)
        mp_matches = data_mp["certain_matches"] + data_mp["probable_matches"]
        assert len(mp_matches) >= 1



from __future__ import annotations

from rapidfuzz.distance import JaroWinkler

from gedcom_tools.commands.compare.models import CompareIndividual, MatchScore

_WEIGHTS: dict[str, float] = {
    "Surname": 0.30,
    "Given Name": 0.20,
    "Birth Year": 0.20,
    "Death Year": 0.10,
    "Birth Place": 0.10,
    "Death Place": 0.05,
    "Sex": 0.05,
}

_CORROBORATING_FIELDS: set[str] = {
    "Birth Year",
    "Death Year",
    "Birth Place",
    "Death Place",
}

_YEAR_BANDS: list[tuple[int, float]] = [
    (0, 1.0),
    (1, 0.85),
    (2, 0.70),
    (3, 0.50),
    (5, 0.25),
    (7, 0.10),
    (10, 0.05),
]


def _year_proximity(year_a: int, year_b: int) -> float:
    diff = abs(year_a - year_b)
    for threshold, score in _YEAR_BANDS:
        if diff <= threshold:
            return score
    return 0.0


def _best_name_jw(
    primary_a: str,
    alts_a: list[str],
    primary_b: str,
    alts_b: list[str],
) -> float:
    candidates_a = [s for s in [primary_a] + alts_a if s]
    candidates_b = [s for s in [primary_b] + alts_b if s]
    if not candidates_a or not candidates_b:
        return 0.0
    best = 0.0
    for ca in candidates_a:
        for cb in candidates_b:
            sim = JaroWinkler.similarity(ca, cb)
            if sim > best:
                best = sim
    return best


def _place_similarity(place_a: str, place_b: str) -> float:
    parts_a = [p.strip() for p in place_a.split(",") if p.strip()]
    parts_b = [p.strip() for p in place_b.split(",") if p.strip()]
    if not parts_a or not parts_b:
        return 0.0

    if len(parts_a) <= len(parts_b):
        shorter, longer = parts_a, list(parts_b)
    else:
        shorter, longer = parts_b, list(parts_a)

    total = 0.0
    for comp in shorter:
        best_score = 0.0
        best_idx = 0
        for i, candidate in enumerate(longer):
            sim = JaroWinkler.similarity(comp, candidate)
            if sim > best_score:
                best_score = sim
                best_idx = i
        total += best_score
        longer.pop(best_idx)

    return total / len(shorter)


def score_pair(
    a: CompareIndividual,
    b: CompareIndividual,
    certain_threshold: float = 0.85,
    probable_threshold: float = 0.65,
    reject_sex_mismatch: bool = False,
) -> MatchScore:
    """Score a candidate pair and classify the match."""
    field_scores: dict[str, float] = {}
    applicable_weights: dict[str, float] = {}
    sex_penalty = False

    # Surname
    from gedcom_tools.phonetics import phonetic_codes_match

    candidates_a = [s for s in [a.surname_normalized] + a.alt_surnames_normalized if s]
    candidates_b = [s for s in [b.surname_normalized] + b.alt_surnames_normalized if s]
    if candidates_a and candidates_b:
        jw = _best_name_jw(
            a.surname_normalized,
            a.alt_surnames_normalized,
            b.surname_normalized,
            b.alt_surnames_normalized,
        )
        if (
            phonetic_codes_match(
                a.surname_phonetic,
                a.surname_phonetic_alt,
                b.surname_phonetic,
                b.surname_phonetic_alt,
            )
            and 0.50 <= jw <= 0.85
        ):
            jw = min(jw + 0.05, 1.0)
        field_scores["Surname"] = jw
        applicable_weights["Surname"] = _WEIGHTS["Surname"]

    # Given Name
    candidates_a = [
        s for s in [a.given_name_normalized] + a.alt_given_names_normalized if s
    ]
    candidates_b = [
        s for s in [b.given_name_normalized] + b.alt_given_names_normalized if s
    ]
    if candidates_a and candidates_b:
        field_scores["Given Name"] = _best_name_jw(
            a.given_name_normalized,
            a.alt_given_names_normalized,
            b.given_name_normalized,
            b.alt_given_names_normalized,
        )
        applicable_weights["Given Name"] = _WEIGHTS["Given Name"]

    # Birth Year
    if a.birth_year is not None and b.birth_year is not None:
        field_scores["Birth Year"] = _year_proximity(a.birth_year, b.birth_year)
        applicable_weights["Birth Year"] = _WEIGHTS["Birth Year"]

    # Death Year
    if a.death_year is not None and b.death_year is not None:
        field_scores["Death Year"] = _year_proximity(a.death_year, b.death_year)
        applicable_weights["Death Year"] = _WEIGHTS["Death Year"]

    # Birth Place
    if a.birth_place_normalized and b.birth_place_normalized:
        field_scores["Birth Place"] = _place_similarity(
            a.birth_place_normalized, b.birth_place_normalized
        )
        applicable_weights["Birth Place"] = _WEIGHTS["Birth Place"]

    # Death Place
    if a.death_place_normalized and b.death_place_normalized:
        field_scores["Death Place"] = _place_similarity(
            a.death_place_normalized, b.death_place_normalized
        )
        applicable_weights["Death Place"] = _WEIGHTS["Death Place"]

    # Sex
    if a.sex and b.sex:
        if a.sex == b.sex:
            field_scores["Sex"] = 1.0
            applicable_weights["Sex"] = _WEIGHTS["Sex"]
        else:
            if reject_sex_mismatch:
                return MatchScore(
                    total=0.0,
                    field_scores={},
                    classification="non_match",
                    comparable_field_count=0,
                )
            sex_penalty = True

    # Counts
    comparable_count = len(field_scores)
    corroborating_count = sum(1 for f in field_scores if f in _CORROBORATING_FIELDS)
    name_only = corroborating_count == 0 and comparable_count > 0
    # Insufficient when we lack either breadth (<3 fields) or depth (no
    # corroborating date/place fields).  Can coexist with "probable" --
    # e.g. name+sex match with high JW but no dates or places.
    # Consumers should treat insufficient_data + probable as low-confidence.
    insufficient = comparable_count < 3 or corroborating_count == 0

    # Weighted average
    if not applicable_weights:
        total = 0.0
    else:
        weight_sum = sum(applicable_weights.values())
        total = (
            sum(field_scores[f] * applicable_weights[f] for f in field_scores)
            / weight_sum
        )

    # Sex penalty
    if sex_penalty:
        total *= 0.7

    # Classification
    if comparable_count < 3:
        classification = "non_match"
    elif name_only:
        classification = "probable" if total >= probable_threshold else "non_match"
    elif comparable_count < 4:
        classification = "probable" if total >= probable_threshold else "non_match"
    else:
        if total >= certain_threshold:
            classification = "certain"
        elif total >= probable_threshold:
            classification = "probable"
        else:
            classification = "non_match"

    return MatchScore(
        total=round(total, 4),
        field_scores={k: round(v, 4) for k, v in field_scores.items()},
        classification=classification,
        insufficient_data=insufficient,
        name_only=name_only,
        comparable_field_count=comparable_count,
        sex_penalty=sex_penalty,
    )



"""Data models for the filter command."""

from __future__ import annotations

import json
from dataclasses import dataclass, field

from gedcom_tools import __version__
from gedcom_tools.progress import Colors

UNLIMITED_DEPTH: int = 2**20


@dataclass
class GedcomLine:
    level: int
    xref: str | None
    tag: str
    value: str | None
    raw: str
    line_number: int


@dataclass
class GedcomRecord:
    header: GedcomLine
    children: list[GedcomLine]

    @property
    def xref(self) -> str | None:
        return self.header.xref

    @property
    def tag(self) -> str:
        return self.header.tag


@dataclass
class RecordCounts:
    indi: int = 0
    fam: int = 0
    note: int = 0
    sour: int = 0
    obje: int = 0
    repo: int = 0
    subm: int = 0
    other: int = 0

    @property
    def total(self) -> int:
        return (
            self.indi
            + self.fam
            + self.note
            + self.sour
            + self.obje
            + self.repo
            + self.subm
            + self.other
        )


@dataclass
class FilterSpec:
    strip_custom_tags: bool = False
    strip_notes: bool = False
    strip_sources: bool = False
    strip_multimedia: bool = False
    strip_tags: list[str] = field(default_factory=list)
    subtree_root: str | None = None
    ancestor_depth: int | None = None
    descendant_depth: int = 0
    include_spouses: bool = False


@dataclass
class FilterResult:
    source_path: str
    output_path: str
    source_counts: RecordCounts
    output_counts: RecordCounts
    removed_counts: RecordCounts
    dangling_lines_removed: int
    empty_families_removed: int
    dry_run: bool

    def format_text(self, colors: Colors, quiet: bool) -> str:
        if quiet:
            line = (
                f"Filtered {self.source_path} "
                f"({self.source_counts.total:,} \u2192 "
                f"{self.output_counts.total:,} records) "
                f"\u2192 {self.output_path}"
            )
            if self.dry_run:
                line += " (dry run)"
            return line

        rows: list[tuple[str, int, int, int]] = []
        for label, src, out, rem in [
            (
                "Individuals",
                self.source_counts.indi,
                self.output_counts.indi,
                self.removed_counts.indi,
            ),
            (
                "Families",
                self.source_counts.fam,
                self.output_counts.fam,
                self.removed_counts.fam,
            ),
            (
                "Notes",
                self.source_counts.note,
                self.output_counts.note,
                self.removed_counts.note,
            ),
            (
                "Sources",
                self.source_counts.sour,
                self.output_counts.sour,
                self.removed_counts.sour,
            ),
            (
                "Multimedia",
                self.source_counts.obje,
                self.output_counts.obje,
                self.removed_counts.obje,
            ),
            (
                "Repositories",
                self.source_counts.repo,
                self.output_counts.repo,
                self.removed_counts.repo,
            ),
            (
                "Submitters",
                self.source_counts.subm,
                self.output_counts.subm,
                self.removed_counts.subm,
            ),
            (
                "Other",
                self.source_counts.other,
                self.output_counts.other,
                self.removed_counts.other,
            ),
        ]:
            if src > 0 or out > 0 or rem > 0:
                rows.append((label, src, out, rem))

        lines: list[str] = [
            f"File: {self.source_path}",
            "",
            f"{colors.cyan}=== Filter Results ==={colors.reset}",
            "",
            f"  {'Record Type':<15} {'Source':>8} {'Output':>8} {'Removed':>8}",
            f"  {'-' * 15} {'-' * 8} {'-' * 8} {'-' * 8}",
        ]

        for label, src, out, rem in rows:
            lines.append(f"  {label:<15} {src:>8,} {out:>8,} {rem:>8,}")

        lines.append(f"  {'-' * 15} {'-' * 8} {'-' * 8} {'-' * 8}")
        lines.append(
            f"  {'Total':<15} {self.source_counts.total:>8,} "
            f"{self.output_counts.total:>8,} {self.removed_counts.total:>8,}"
        )

        if self.dangling_lines_removed > 0:
            lines.append(
                f"\n  Dangling references cleaned: {self.dangling_lines_removed:,}"
            )
        if self.empty_families_removed > 0:
            lines.append(
                f"  Empty families removed:      {self.empty_families_removed:,}"
            )

        lines.append(f"\n  Output: {self.output_path}")

        if self.dry_run:
            lines.append("\n  (dry run \u2014 no file written)")

        return "\n".join(lines)

    def format_json(self) -> str:
        def _counts_dict(c: RecordCounts) -> dict[str, int]:
            return {
                "individuals": c.indi,
                "families": c.fam,
                "notes": c.note,
                "sources": c.sour,
                "multimedia": c.obje,
                "repositories": c.repo,
                "submitters": c.subm,
                "other": c.other,
                "total": c.total,
            }

        data = {
            "source_file": self.source_path,
            "output_file": self.output_path,
            "source": _counts_dict(self.source_counts),
            "output": _counts_dict(self.output_counts),
            "removed": _counts_dict(self.removed_counts),
            "dangling_lines_removed": self.dangling_lines_removed,
            "empty_families_removed": self.empty_families_removed,
            "dry_run": self.dry_run,
            "gedcom_tools_version": __version__,
        }
        return json.dumps(data, indent=2, ensure_ascii=False)



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from gedcom_tools.commands.compare.models import (
        CompareIndividual,
        MatchPair,
    )
    from gedcom_tools.commands.duplicates.models import DuplicatesResult
    from gedcom_tools.progress import Colors


def _year_range(birth: int | None, death: int | None) -> str:
    b = str(birth) if birth is not None else "?"
    d = str(death) if death is not None else "?"
    return f"{b}-{d}"


def _format_match_section(
    pairs: list[MatchPair],
    section_title: str,
    colors: Colors,
    verbose: bool,
    limit: int,
) -> list[str]:
    total = len(pairs)
    if total == 0:
        return []

    lines: list[str] = []

    if limit > 0 and total > limit:
        header = f"=== {section_title} ({total} total, showing first {limit}) ==="
        display = pairs[:limit]
    else:
        header = f"=== {section_title} ({total}) ==="
        display = pairs

    lines.append(f"{colors.cyan}{header}{colors.reset}")

    for pair in display:
        a = pair.individual_a
        b = pair.individual_b
        a_years = _year_range(a.birth_year, a.death_year)
        b_years = _year_range(b.birth_year, b.death_year)
        a_name = a.full_name or "Unknown"
        b_name = b.full_name or "Unknown"
        suffix = ""
        if pair.score.insufficient_data:
            suffix = " (low confidence)"
        lines.append(
            f"  {a_name} ({a_years}) [{a.xref}] "
            f"\u2194 {b_name} ({b_years}) [{b.xref}]  "
            f"score: {pair.score.total:.2f}{suffix}"
        )

        if pair.field_diffs:
            for diff in pair.field_diffs:
                lines.append(f'    {diff.field}: "{diff.value_a}" vs "{diff.value_b}"')
        else:
            lines.append("    (no differences)")

        if verbose and pair.score.field_scores:
            parts = [f"{k} {v:.2f}" for k, v in pair.score.field_scores.items()]
            if pair.score.sex_penalty:
                parts.append("Sex mismatch \u00d70.70")
            lines.append(f"    [Scores: {', '.join(parts)}]")

    if limit > 0 and total > limit:
        remaining = total - limit
        lines.append(f"  ({remaining} more \u2014 use --limit 0 for all)")

    return lines


def format_text(
    result: DuplicatesResult,
    colors: Colors,
    quiet: bool = False,
    verbose: bool = False,
    show_matches: str = "all",
    limit: int = 0,
) -> str:
    if quiet:
        return (
            f"{len(result.certain_matches)} certain, "
            f"{len(result.probable_matches)} probable"
        )

    lines: list[str] = [f"File: {result.file}"]
    lines.append("")
    lines.append(f"{colors.cyan}=== Duplicate Scan Summary ==={colors.reset}")
    lines.append(f"  Individuals scanned: {result.total_individuals:>5}")
    lines.append(f"  Certain duplicates:  {len(result.certain_matches):>5}")
    lines.append(f"  Probable duplicates: {len(result.probable_matches):>5}")

    if show_matches in ("all", "certain"):
        certain_lines = _format_match_section(
            result.certain_matches, "Certain Duplicates", colors, verbose, limit
        )
        if certain_lines:
            lines.append("")
            lines.extend(certain_lines)

    if show_matches in ("all", "probable"):
        probable_lines = _format_match_section(
            result.probable_matches, "Probable Duplicates", colors, verbose, limit
        )
        if probable_lines:
            lines.append("")
            lines.extend(probable_lines)

    return "\n".join(lines)


def _individual_to_dict(ind: CompareIndividual) -> dict[str, Any]:
    return {
        "xref": ind.xref,
        "name": ind.full_name or "Unknown",
        "given_name": ind.given_name,
        "surname": ind.surname,
        "sex": ind.sex,
        "birth_year": ind.birth_year,
        "birth_place": ind.birth_place,
        "death_year": ind.death_year,
        "death_place": ind.death_place,
    }


def _pair_to_dict(pair: MatchPair) -> dict[str, Any]:
    d: dict[str, Any] = {
        "individual_a": _individual_to_dict(pair.individual_a),
        "individual_b": _individual_to_dict(pair.individual_b),
        "score": round(pair.score.total, 4),
        "classification": pair.score.classification,
        "field_scores": {k: round(v, 4) for k, v in pair.score.field_scores.items()},
        "differences": [
            {
                "field": fd.field,
                "value_a": fd.value_a,
                "value_b": fd.value_b,
            }
            for fd in pair.field_diffs
        ],
    }
    if pair.score.insufficient_data:
        d["insufficient_data"] = True
    return d


def format_json(
    result: DuplicatesResult,
    show_matches: str = "all",
    limit: int = 0,
) -> str:
    certain = result.certain_matches if show_matches in ("all", "certain") else []
    probable = result.probable_matches if show_matches in ("all", "probable") else []

    certain_total = len(certain)
    probable_total = len(probable)

    certain_display = certain[:limit] if limit > 0 else certain
    probable_display = probable[:limit] if limit > 0 else probable

    output: dict[str, Any] = {
        "file": result.file,
        "encoding": {
            "detected": result.encoding.encoding,
            "has_bom": result.encoding.has_bom,
            "declared": result.encoding.declared_charset,
        },
        "total_individuals": result.total_individuals,
        "certain_duplicates": [_pair_to_dict(p) for p in certain_display],
        "certain_duplicates_total": certain_total,
        "probable_duplicates": [_pair_to_dict(p) for p in probable_display],
        "probable_duplicates_total": probable_total,
    }

    return json.dumps(output, indent=2, ensure_ascii=False)



"""Validation issue types and error codes."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum


class Severity(Enum):
    """Severity level of a validation issue."""

    ERROR = "error"
    WARNING = "warning"


class ErrorCode(Enum):
    """Error and warning codes for validation issues.

    E0xx = Errors (fatal issues that make the file invalid)
    W0xx = Warnings (issues that should be reviewed but don't invalidate the file)
    """

    # Errors: Structure and parsing
    E001_UNRESOLVED_XREF = "E001"
    E002_DUPLICATE_XREF = "E002"
    E003_INVALID_LEVEL = "E003"
    E004_MALFORMED_LINE = "E004"
    E005_MISSING_HEAD = "E005"
    E006_MISSING_TRLR = "E006"
    E007_CONTENT_AFTER_TRLR = "E007"
    E008_DECODE_FAILURE = "E008"
    # Errors: Semantic
    E010_ANCESTRY_CYCLE = "E010"
    E011_DEATH_BEFORE_BIRTH = "E011"
    E012_BIRTH_BEFORE_PARENT = "E012"

    # Errors: Version compliance (strict mode)
    E013_MISSING_GEDC = "E013"
    E014_MISSING_GEDC_VERS = "E014"
    E015_MISSING_SOUR = "E015"
    E016_MISSING_CHAR = "E016"

    # Warnings: Structure
    W002_TRAILING_WHITESPACE = "W002"
    W003_LINE_TOO_LONG = "W003"
    W004_CUSTOM_TAG = "W004"
    W005_MISSING_SUBM = "W005"

    # Warnings: References
    W010_ORPHANED_NOTE = "W010"
    W011_ORPHANED_OBJE = "W011"
    W012_ORPHANED_SOUR = "W012"
    W013_ORPHANED_REPO = "W013"
    W014_ISOLATED_INDI = "W014"
    W015_EMPTY_FAM = "W015"
    W016_ASYMMETRIC_CHILD_LINK = "W016"
    W017_ASYMMETRIC_SPOUSE_LINK = "W017"

    # Warnings: Semantic
    W020_PARENT_TOO_YOUNG = "W020"
    W021_MOTHER_TOO_OLD = "W021"
    W022_FATHER_TOO_OLD = "W022"
    W023_AGE_AT_DEATH_IMPLAUSIBLE = "W023"
    W024_MARRIAGE_BEFORE_BIRTH = "W024"
    W025_CHILD_BEFORE_MARRIAGE = "W025"
    W026_SIBLING_TOO_CLOSE = "W026"
    W027_MULTIPLE_SEX = "W027"
    W028_INVALID_SEX = "W028"
    W029_SEX_ROLE_MISMATCH = "W029"

    # Warnings: Version compliance (strict mode)
    W030_ANSEL_DEPRECATED = "W030"
    W031_VERSION_MISMATCH = "W031"
    W032_LINE_TOO_LONG_STRICT = "W032"
    W033_OBJE_MISSING_FILE = "W033"
    W034_FILE_MISSING_FORM = "W034"

    @property
    def severity(self) -> Severity:
        """Return the severity based on code prefix."""
        return Severity.ERROR if self.value.startswith("E") else Severity.WARNING

    @property
    def description(self) -> str:
        """Return a human-readable description of the error code."""
        descriptions = {
            # Errors
            "E001": "Unresolved cross-reference",
            "E002": "Duplicate cross-reference definition",
            "E003": "Invalid level number",
            "E004": "Malformed GEDCOM line",
            "E005": "Missing HEAD record",
            "E006": "Missing TRLR record",
            "E007": "Content after TRLR record",
            "E008": "Character encoding decode failure",
            "E010": "Ancestry cycle detected",
            "E011": "Death date before birth date",
            "E012": "Birth date before parent's birth",
            "E013": "Missing GEDC in HEAD",
            "E014": "Missing VERS in GEDC",
            "E015": "Missing SOUR in HEAD",
            "E016": "Missing CHAR in HEAD",
            # Warnings
            "W002": "Trailing whitespace",
            "W003": "Line exceeds recommended length",
            "W004": "Custom/non-standard tag",
            "W005": "Missing SUBM record",
            "W010": "Orphaned NOTE record",
            "W011": "Orphaned OBJE record",
            "W012": "Orphaned SOUR record",
            "W013": "Orphaned REPO record",
            "W014": "Individual with no family connections",
            "W015": "Family with no members",
            "W016": "Family-child cross-reference exists on only one side",
            "W017": "Family-spouse cross-reference exists on only one side",
            "W020": "Parent too young at child's birth",
            "W021": "Mother too old at child's birth",
            "W022": "Father too old at child's birth",
            "W023": "Implausible age at death",
            "W024": "Marriage before birth",
            "W025": "Child born before parents' marriage",
            "W026": "Siblings born less than 9 months apart",
            "W027": "Individual has multiple SEX records",
            "W028": "Invalid SEX value",
            "W029": "SEX conflicts with spousal role",
            "W030": "ANSEL encoding deprecated in GEDCOM 5.5.5",
            "W031": "Declared version does not match --strict version",
            "W032": "Line exceeds 255 byte limit (strict)",
            "W033": "OBJE record has no FILE subtag",
            "W034": "FILE subtag has no FORM",
        }
        return descriptions.get(self.value, "Unknown issue")


@dataclass
class ValidationIssue:
    """A single validation issue found in a GEDCOM file."""

    code: ErrorCode
    message: str
    line: int | None = None
    xref: str | None = None
    context: str | None = None

    @property
    def severity(self) -> Severity:
        """Return the severity of this issue."""
        return self.code.severity

    def __str__(self) -> str:
        """Return a human-readable representation."""
        parts = [f"[{self.code.value}]"]
        if self.line is not None:
            parts.append(f"Line {self.line}:")
        if self.xref:
            parts.append(f"({self.xref})")
        parts.append(self.message)
        return " ".join(parts)


@dataclass
class IndividualInfo:
    """Collected information about an individual for semantic validation."""

    xref: str
    line: int
    birth_year: int | None = None
    birth_month: int | None = None
    death_year: int | None = None
    sex: str | None = None
    famc_xrefs: list[str] = field(default_factory=list)
    fams_xrefs: list[str] = field(default_factory=list)


@dataclass
class FamilyInfo:
    """Collected information about a family for semantic validation."""

    xref: str
    line: int
    husb_xref: str | None = None
    wife_xref: str | None = None
    chil_xrefs: list[str] = field(default_factory=list)
    marriage_year: int | None = None


@dataclass
class RecordInfo:
    """Basic information about any record."""

    xref: str
    record_type: str
    line: int


@dataclass
class UsageInfo:
    """Information about where an xref is used."""

    line: int
    context: str
